# coding=utf-8
# Copyright 2019-present, Facebook, Inc and the HuggingFace Inc. team.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
""" PyTorch XLM model.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import logging
import math

import itertools
import numpy as np

import torch
from torch import nn
from torch.nn import functional as F
from torch.nn import CrossEntropyLoss, MSELoss


from relogic.logickit.inference.modeling_utils import prune_linear_layer, PreTrainedModel, SequenceSummary, SQuADHead
from relogic.logickit.base.configuration_xlm import XLMConfig
from relogic.utils.file_utils import add_start_docstrings

logger = logging.getLogger(__name__)

XLM_PRETRAINED_MODEL_ARCHIVE_MAP = {
    'xlm-mlm-en-2048': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-en-2048-pytorch_model.bin",
    'xlm-mlm-ende-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-ende-1024-pytorch_model.bin",
    'xlm-mlm-enfr-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-enfr-1024-pytorch_model.bin",
    'xlm-mlm-enro-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-enro-1024-pytorch_model.bin",
    'xlm-mlm-tlm-xnli15-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-tlm-xnli15-1024-pytorch_model.bin",
    'xlm-mlm-xnli15-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-xnli15-1024-pytorch_model.bin",
    'xlm-clm-enfr-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-clm-enfr-1024-pytorch_model.bin",
    'xlm-clm-ende-1024': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-clm-ende-1024-pytorch_model.bin",
    'xlm-mlm-17-1280': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-17-1280-pytorch_model.bin",
    'xlm-mlm-100-1280': "https://s3.amazonaws.com/models.huggingface.co/bert/xlm-mlm-100-1280-pytorch_model.bin",
}


def create_sinusoidal_embeddings(n_pos, dim, out):
    position_enc = np.array([
        [pos / np.power(10000, 2 * (j // 2) / dim) for j in range(dim)]
        for pos in range(n_pos)
    ])
    out[:, 0::2] = torch.FloatTensor(np.sin(position_enc[:, 0::2]))
    out[:, 1::2] = torch.FloatTensor(np.cos(position_enc[:, 1::2]))
    out.detach_()
    out.requires_grad = False


def gelu(x):
    """
    GELU activation
    https://arxiv.org/abs/1606.08415
    https://github.com/huggingface/pytorch-openai-transformer-lm/blob/master/model_pytorch.py#L14
    https://github.com/huggingface/transformers/blob/master/modeling.py
    """
    # return 0.5 * x * (1 + torch.tanh(math.sqrt(2 / math.pi) * (x + 0.044715 * torch.pow(x, 3))))
    return 0.5 * x * (1.0 + torch.erf(x / math.sqrt(2.0)))


def get_masks(slen, lengths, causal, padding_mask=None):
    """
    Generate hidden states mask, and optionally an attention mask.
    """
    bs = lengths.size(0)
    if padding_mask is not None:
        mask = padding_mask
    else:
        assert lengths.max().item() <= slen
        alen = torch.arange(slen, dtype=torch.long, device=lengths.device)
        mask = alen < lengths[:, None]

    # attention mask is the same as mask, or triangular inferior attention (causal)
    if causal:
        attn_mask = alen[None, None, :].repeat(bs, slen, 1) <= alen[None, :, None]
    else:
        attn_mask = mask

    # sanity check
    assert mask.size() == (bs, slen)
    assert causal is False or attn_mask.size() == (bs, slen, slen)

    return mask, attn_mask


class MultiHeadAttention(nn.Module):

    NEW_ID = itertools.count()

    def __init__(self, n_heads, dim, config):
        super(MultiHeadAttention, self).__init__()
        self.layer_id = next(MultiHeadAttention.NEW_ID)
        self.output_attentions = config.output_attentions
        self.dim = dim
        self.n_heads = n_heads
        self.dropout = config.attention_dropout
        assert self.dim % self.n_heads == 0

        self.q_lin = nn.Linear(dim, dim)
        self.k_lin = nn.Linear(dim, dim)
        self.v_lin = nn.Linear(dim, dim)
        self.out_lin = nn.Linear(dim, dim)
        self.pruned_heads = set()

    def prune_heads(self, heads):
        attention_head_size = self.dim // self.n_heads
        if len(heads) == 0:
            return
        mask = torch.ones(self.n_heads, attention_head_size)
        heads = set(heads) - self.pruned_heads
        for head in heads:
            head -= sum(1 if h < head else 0 for h in self.pruned_heads)
            mask[head] = 0
        mask = mask.view(-1).contiguous().eq(1)
        index = torch.arange(len(mask))[mask].long()
        # Prune linear layers
        self.q_lin = prune_linear_layer(self.q_lin, index)
        self.k_lin = prune_linear_layer(self.k_lin, index)
        self.v_lin = prune_linear_layer(self.v_lin, index)
        self.out_lin = prune_linear_layer(self.out_lin, index, dim=1)
        # Update hyper params
        self.n_heads = self.n_heads - len(heads)
        self.dim = attention_head_size * self.n_heads
        self.pruned_heads = self.pruned_heads.union(heads)

    def forward(self, input, mask, kv=None, cache=None, head_mask=None):
        """
        Self-attention (if kv is None) or attention over source sentence (provided by kv).
        """
        # Input is (bs, qlen, dim)
        # Mask is (bs, klen) (non-causal) or (bs, klen, klen)
        bs, qlen, dim = input.size()
        if kv is None:
            klen = qlen if cache is None else cache['slen'] + qlen
        else:
            klen = kv.size(1)
        # assert dim == self.dim, 'Dimensions do not match: %s input vs %s configured' % (dim, self.dim)
        n_heads = self.n_heads
        dim_per_head = self.dim // n_heads
        mask_reshape = (bs, 1, qlen, klen) if mask.dim() == 3 else (bs, 1, 1, klen)

        def shape(x):
            """  projection """
            return x.view(bs, -1, self.n_heads, dim_per_head).transpose(1, 2)

        def unshape(x):
            """  compute context """
            return x.transpose(1, 2).contiguous().view(bs, -1, self.n_heads * dim_per_head)

        q = shape(self.q_lin(input))                                          # (bs, n_heads, qlen, dim_per_head)
        if kv is None:
            k = shape(self.k_lin(input))                                      # (bs, n_heads, qlen, dim_per_head)
            v = shape(self.v_lin(input))                                      # (bs, n_heads, qlen, dim_per_head)
        elif cache is None or self.layer_id not in cache:
            k = v = kv
            k = shape(self.k_lin(k))                                          # (bs, n_heads, qlen, dim_per_head)
            v = shape(self.v_lin(v))                                          # (bs, n_heads, qlen, dim_per_head)

        if cache is not None:
            if self.layer_id in cache:
                if kv is None:
                    k_, v_ = cache[self.layer_id]
                    k = torch.cat([k_, k], dim=2)                             # (bs, n_heads, klen, dim_per_head)
                    v = torch.cat([v_, v], dim=2)                             # (bs, n_heads, klen, dim_per_head)
                else:
                    k, v = cache[self.layer_id]
            cache[self.layer_id] = (k, v)

        q = q / math.sqrt(dim_per_head)                                       # (bs, n_heads, qlen, dim_per_head)
        scores = torch.matmul(q, k.transpose(2, 3))                           # (bs, n_heads, qlen, klen)
        mask = (mask == 0).view(mask_reshape).expand_as(scores)               # (bs, n_heads, qlen, klen)
        scores.masked_fill_(mask, -float('inf'))                              # (bs, n_heads, qlen, klen)

        weights = F.softmax(scores.float(), dim=-1).type_as(scores)           # (bs, n_heads, qlen, klen)
        weights = F.dropout(weights, p=self.dropout, training=self.training)  # (bs, n_heads, qlen, klen)

        # Mask heads if we want to
        if head_mask is not None:
            weights = weights * head_mask

        context = torch.matmul(weights, v)                                    # (bs, n_heads, qlen, dim_per_head)
        context = unshape(context)                                            # (bs, qlen, dim)

        outputs = (self.out_lin(context),)
        if self.output_attentions:
            outputs = outputs + (weights,)
        return outputs


class TransformerFFN(nn.Module):

    def __init__(self, in_dim, dim_hidden, out_dim, config):
        super(TransformerFFN, self).__init__()
        self.dropout = config.dropout
        self.lin1 = nn.Linear(in_dim, dim_hidden)
        self.lin2 = nn.Linear(dim_hidden, out_dim)
        self.act = gelu if config.gelu_activation else F.relu

    def forward(self, input):
        x = self.lin1(input)
        x = self.act(x)
        x = self.lin2(x)
        x = F.dropout(x, p=self.dropout, training=self.training)
        return x


class XLMPreTrainedModel(PreTrainedModel):
    """ An abstract class to handle weights initialization and
        a simple interface for dowloading and loading pretrained models.
    """
    config_class = XLMConfig
    pretrained_model_archive_map = XLM_PRETRAINED_MODEL_ARCHIVE_MAP
    load_tf_weights = None
    base_model_prefix = "transformer"

    def __init__(self, *inputs, **kwargs):
        super(XLMPreTrainedModel, self).__init__(*inputs, **kwargs)

    def _init_weights(self, module):
        """ Initialize the weights. """
        if isinstance(module, nn.Embedding):
            if self.config is not None and self.config.embed_init_std is not None:
                nn.init.normal_(module.weight, mean=0, std=self.config.embed_init_std)
        if isinstance(module, nn.Linear):
            if self.config is not None and self.config.init_std is not None:
                nn.init.normal_(module.weight, mean=0, std=self.config.init_std)
                if hasattr(module, 'bias') and module.bias is not None:
                    nn.init.constant_(module.bias, 0.)
        if isinstance(module, nn.LayerNorm):
            module.bias.data.zero_()
            module.weight.data.fill_(1.0)


XLM_START_DOCSTRING = r"""    The XLM model was proposed in
    `Cross-lingual Language Model Pretraining`_
    by Guillaume Lample*, Alexis Conneau*. It's a transformer pre-trained using one of the following objectives:

        - a causal language modeling (CLM) objective (next token prediction),
        - a masked language modeling (MLM) objective (Bert-like), or
        - a Translation Language Modeling (TLM) object (extension of Bert's MLM to multiple language inputs)

    Original code can be found `here`_.

    This model is a PyTorch `torch.nn.Module`_ sub-class. Use it as a regular PyTorch Module and
    refer to the PyTorch documentation for all matter related to general usage and behavior.

    .. _`Cross-lingual Language Model Pretraining`:
        https://arxiv.org/abs/1901.07291

    .. _`torch.nn.Module`:
        https://pytorch.org/docs/stable/nn.html#module

    .. _`here`:
        https://github.com/facebookresearch/XLM

    Parameters:
        config (:class:`~transformers.XLMConfig`): Model configuration class with all the parameters of the model.
            Initializing with a config file does not load the weights associated with the model, only the configuration.
            Check out the :meth:`~transformers.PreTrainedModel.from_pretrained` method to load the model weights.
"""

XLM_INPUTS_DOCSTRING = r"""
    Inputs:
        **input_ids**: ``torch.LongTensor`` of shape ``(batch_size, sequence_length)``:
            Indices of input sequence tokens in the vocabulary.

            XLM is a model with absolute position embeddings so it's usually advised to pad the inputs on
            the right rather than the left.

            Indices can be obtained using :class:`transformers.XLMTokenizer`.
            See :func:`transformers.PreTrainedTokenizer.encode` and
            :func:`transformers.PreTrainedTokenizer.convert_tokens_to_ids` for details.
        **attention_mask**: (`optional`) ``torch.FloatTensor`` of shape ``(batch_size, sequence_length)``:
            Mask to avoid performing attention on padding token indices.
            Mask values selected in ``[0, 1]``:
            ``1`` for tokens that are NOT MASKED, ``0`` for MASKED tokens.
        **langs**: (`optional`) ``torch.LongTensor`` of shape ``(batch_size, sequence_length)``:
            A parallel sequence of tokens to be used to indicate the language of each token in the input.
            Indices are languages ids which can be obtained from the language names by using two conversion mappings
            provided in the configuration of the model (only provided for multilingual models).
            More precisely, the `language name -> language id` mapping is in `model.config.lang2id` (dict str -> int) and
            the `language id -> language name` mapping is `model.config.id2lang` (dict int -> str).
        **token_type_ids**: (`optional`) ``torch.LongTensor`` of shape ``(batch_size, sequence_length)``:
            A parallel sequence of tokens (can be used to indicate various portions of the inputs).
            The embeddings from these tokens will be summed with the respective token embeddings.
            Indices are selected in the vocabulary (unlike BERT which has a specific vocabulary for segment indices).
        **position_ids**: (`optional`) ``torch.LongTensor`` of shape ``(batch_size, sequence_length)``:
            Indices of positions of each input sequence tokens in the position embeddings.
            Selected in the range ``[0, config.max_position_embeddings - 1]``.
        **lengths**: (`optional`) ``torch.LongTensor`` of shape ``(batch_size,)``:
            Length of each sentence that can be used to avoid performing attention on padding token indices.
            You can also use `attention_mask` for the same result (see above), kept here for compatbility.
            Indices selected in ``[0, ..., input_ids.size(-1)]``:
        **cache**:
            dictionary with ``torch.FloatTensor`` that contains pre-computed
            hidden-states (key and values in the attention blocks) as computed by the model
            (see `cache` output below). Can be used to speed up sequential decoding.
            The dictionary object will be modified in-place during the forward pass to add newly computed hidden-states.
        **head_mask**: (`optional`) ``torch.FloatTensor`` of shape ``(num_heads,)`` or ``(num_layers, num_heads)``:
            Mask to nullify selected heads of the self-attention modules.
            Mask values selected in ``[0, 1]``:
            ``1`` indicates the head is **not masked**, ``0`` indicates the head is **masked**.
"""

@add_start_docstrings("The bare XLM Model transformer outputting raw hidden-states without any specific head on top.",
                      XLM_START_DOCSTRING, XLM_INPUTS_DOCSTRING)
class XLMModel(XLMPreTrainedModel):
    r"""
    Outputs: `Tuple` comprising various elements depending on the configuration (config) and inputs:
        **last_hidden_state**: ``torch.FloatTensor`` of shape ``(batch_size, sequence_length, hidden_size)``
            Sequence of hidden-states at the last layer of the model.
        **hidden_states**: (`optional`, returned when ``config.output_hidden_states=True``)
            list of ``torch.FloatTensor`` (one for the output of each layer + the output of the embeddings)
            of shape ``(batch_size, sequence_length, hidden_size)``:
            Hidden-states of the model at the output of each layer plus the initial embedding outputs.
        **attentions**: (`optional`, returned when ``config.output_attentions=True``)
            list of ``torch.FloatTensor`` (one for each layer) of shape ``(batch_size, num_heads, sequence_length, sequence_length)``:
            Attentions weights after the attention softmax, used to compute the weighted average in the self-attention heads.

    Examples::

        tokenizer = XLMTokenizer.from_pretrained('xlm-mlm-en-2048')
        model = XLMModel.from_pretrained('xlm-mlm-en-2048')
        input_ids = torch.tensor(tokenizer.encode("Hello, my dog is cute")).unsqueeze(0)  # Batch size 1
        outputs = model(input_ids)
        last_hidden_states = outputs[0]  # The last hidden-state is the first element of the output tuple

    """
    def __init__(self, config):  #, dico, is_encoder, with_output):
        super(XLMModel, self).__init__(config)
        self.output_attentions = config.output_attentions
        self.output_hidden_states = config.output_hidden_states

        # encoder / decoder, output layer
        self.is_encoder = config.is_encoder
        self.is_decoder = not config.is_encoder
        if self.is_decoder:
            raise NotImplementedError("Currently XLM can only be used as an encoder")
        # self.with_output = with_output
        self.causal = config.causal

        # dictionary / languages
        self.n_langs = config.n_langs
        self.use_lang_emb = config.use_lang_emb
        self.n_words = config.n_words
        self.eos_index = config.eos_index
        self.pad_index = config.pad_index
        # self.dico = dico
        # self.id2lang = config.id2lang
        # self.lang2id = config.lang2id
        # assert len(self.dico) == self.n_words
        # assert len(self.id2lang) == len(self.lang2id) == self.n_langs

        # model parameters
        self.dim = config.emb_dim       # 512 by default
        self.hidden_dim = self.dim * 4  # 2048 by default
        self.n_heads = config.n_heads   # 8 by default
        self.n_layers = config.n_layers
        self.dropout = config.dropout
        self.attention_dropout = config.attention_dropout
        assert self.dim % self.n_heads == 0, 'transformer dim must be a multiple of n_heads'

        # embeddings
        self.position_embeddings = nn.Embedding(config.max_position_embeddings, self.dim)
        if config.sinusoidal_embeddings:
            create_sinusoidal_embeddings(config.max_position_embeddings, self.dim, out=self.position_embeddings.weight)
        if config.n_langs > 1 and config.use_lang_emb:
            self.lang_embeddings = nn.Embedding(self.n_langs, self.dim)
        self.embeddings = nn.Embedding(self.n_words, self.dim, padding_idx=self.pad_index)
        self.layer_norm_emb = nn.LayerNorm(self.dim, eps=config.layer_norm_eps)

        # transformer layers
        self.attentions = nn.ModuleList()
        self.layer_norm1 = nn.ModuleList()
        self.ffns = nn.ModuleList()
        self.layer_norm2 = nn.ModuleList()
        # if self.is_decoder:
        #     self.layer_norm15 = nn.ModuleList()
        #     self.encoder_attn = nn.ModuleList()

        for _ in range(self.n_layers):
            self.attentions.append(MultiHeadAttention(self.n_heads, self.dim, config=config))
            self.layer_norm1.append(nn.LayerNorm(self.dim, eps=config.layer_norm_eps))
            # if self.is_decoder:
            #     self.layer_norm15.append(nn.LayerNorm(self.dim, eps=config.layer_norm_eps))
            #     self.encoder_attn.append(MultiHeadAttention(self.n_heads, self.dim, dropout=self.attention_dropout))
            self.ffns.append(TransformerFFN(self.dim, self.hidden_dim, self.dim, config=config))
            self.layer_norm2.append(nn.LayerNorm(self.dim, eps=config.layer_norm_eps))

        if hasattr(config, "pruned_heads"):
            pruned_heads = config.pruned_heads.copy().items()
            config.pruned_heads = {}
            for layer, heads in pruned_heads:
                if self.attentions[int(layer)].n_heads == config.n_heads:
                    self.prune_heads({int(layer): list(map(int, heads))})

        self.init_weights()

    def _resize_token_embeddings(self, new_num_tokens):
        self.embeddings = self._get_resized_embeddings(self.embeddings, new_num_tokens)
        return self.embeddings

    def _prune_heads(self, heads_to_prune):
        """ Prunes heads of the model.
            heads_to_prune: dict of {layer_num: list of heads to prune in this layer}
            See base class PreTrainedModel
        """
        for layer, heads in heads_to_prune.items():
            self.attentions[layer].prune_heads(heads)

    def forward(self, input_ids, attention_mask=None, langs=None, token_type_ids=None, position_ids=None,
                lengths=None, cache=None, head_mask=None):  # removed: src_enc=None, src_len=None
        if lengths is None:
            lengths = (input_ids != self.pad_index).sum(dim=1).long()
        # mask = input_ids != self.pad_index

        # check inputs
        bs, slen = input_ids.size()
        assert lengths.size(0) == bs
        assert lengths.max().item() <= slen
        # input_ids = input_ids.transpose(0, 1)  # batch size as dimension 0
        # assert (src_enc is None) == (src_len is None)
        # if src_enc is not None:
        #     assert self.is_decoder
        #     assert src_enc.size(0) == bs

        # generate masks
        mask, attn_mask = get_masks(slen, lengths, self.causal, padding_mask=attention_mask)
        # if self.is_decoder and src_enc is not None:
        #     src_mask = torch.arange(src_len.max(), dtype=torch.long, device=lengths.device) < src_len[:, None]

        # position_ids
        if position_ids is None:
            position_ids = input_ids.new((slen,)).long()
            position_ids = torch.arange(slen, out=position_ids).unsqueeze(0)
        else:
            assert position_ids.size() == (bs, slen)  # (slen, bs)
            # position_ids = position_ids.transpose(0, 1)

        # langs
        if langs is not None:
            assert langs.size() == (bs, slen)  # (slen, bs)
            # langs = langs.transpose(0, 1)

        # Prepare head mask if needed
        # 1.0 in head_mask indicate we keep the head
        # attention_probs has shape bsz x n_heads x N x N
        # input head_mask has shape [num_heads] or [num_hidden_layers x num_heads]
        # and head_mask is converted to shape [num_hidden_layers x batch x num_heads x qlen x klen]
        if head_mask is not None:
            if head_mask.dim() == 1:
                head_mask = head_mask.unsqueeze(0).unsqueeze(0).unsqueeze(-1).unsqueeze(-1)
                head_mask = head_mask.expand(self.n_layers, -1, -1, -1, -1)
            elif head_mask.dim() == 2:
                head_mask = head_mask.unsqueeze(1).unsqueeze(-1).unsqueeze(-1)  # We can specify head_mask for each layer
            head_mask = head_mask.to(dtype=next(self.parameters()).dtype) # switch to fload if need + fp16 compatibility
        else:
            head_mask = [None] * self.n_layers

        # do not recompute cached elements
        if cache is not None:
            _slen = slen - cache['slen']
            input_ids = input_ids[:, -_slen:]
            position_ids = position_ids[:, -_slen:]
            if langs is not None:
                langs = langs[:, -_slen:]
            mask = mask[:, -_slen:]
            attn_mask = attn_mask[:, -_slen:]

        # embeddings
        tensor = self.embeddings(input_ids)
        tensor = tensor + self.position_embeddings(position_ids).expand_as(tensor)
        if langs is not None and self.use_lang_emb:
            tensor = tensor + self.lang_embeddings(langs)
        if token_type_ids is not None:
            tensor = tensor + self.embeddings(token_type_ids)
        tensor = self.layer_norm_emb(tensor)
        tensor = F.dropout(tensor, p=self.dropout, training=self.training)
        tensor *= mask.unsqueeze(-1).to(tensor.dtype)

        # transformer layers
        hidden_states = ()
        attentions = ()
        for i in range(self.n_layers):
            if self.output_hidden_states:
                hidden_states = hidden_states + (tensor,)

            # self attention
            attn_outputs = self.attentions[i](tensor, attn_mask, cache=cache, head_mask=head_mask[i])
            attn = attn_outputs[0]
            if self.output_attentions:
                attentions = attentions + (attn_outputs[1],)
            attn = F.dropout(attn, p=self.dropout, training=self.training)
            tensor = tensor + attn
            tensor = self.layer_norm1[i](tensor)

            # encoder attention (for decoder only)
            # if self.is_decoder and src_enc is not None:
            #     attn = self.encoder_attn[i](tensor, src_mask, kv=src_enc, cache=cache)
            #     attn = F.dropout(attn, p=self.dropout, training=self.training)
            #     tensor = tensor + attn
            #     tensor = self.layer_norm15[i](tensor)

            # FFN
            tensor = tensor + self.ffns[i](tensor)
            tensor = self.layer_norm2[i](tensor)
            tensor *= mask.unsqueeze(-1).to(tensor.dtype)

        # Add last hidden state
        if self.output_hidden_states:
            hidden_states = hidden_states + (tensor,)

        # update cache length
        if cache is not None:
            cache['slen'] += tensor.size(1)

        # move back sequence length to dimension 0
        # tensor = tensor.transpose(0, 1)

        outputs = (tensor,)
        if self.output_hidden_states:
            outputs = outputs + (hidden_states,)
        if self.output_attentions:
            outputs = outputs + (attentions,)
        return outputs  # outputs, (hidden_states), (attentions)


class XLMPredLayer(nn.Module):
    """
    Prediction layer (cross_entropy or adaptive_softmax).
    """
    def __init__(self, config):
        super(XLMPredLayer, self).__init__()
        self.asm = config.asm
        self.n_words = config.n_words
        self.pad_index = config.pad_index
        dim = config.emb_dim

        if config.asm is False:
            self.proj = nn.Linear(dim, config.n_words, bias=True)
        else:
            self.proj = nn.AdaptiveLogSoftmaxWithLoss(
                in_features=dim,
                n_classes=config.n_words,
                cutoffs=config.asm_cutoffs,
                div_value=config.asm_div_value,
                head_bias=True,  # default is False
            )

    def forward(self, x, y=None):
        """ Compute the loss, and optionally the scores.
        """
        outputs = ()
        if self.asm is False:
            scores = self.proj(x)
            outputs = (scores,) + outputs
            if y is not None:
                loss = F.cross_entropy(scores.view(-1, self.n_words), y.view(-1), reduction='elementwise_mean')
                outputs = (loss,) + outputs
        else:
            scores = self.proj.log_prob(x)
            outputs = (scores,) + outputs
            if y is not None:
                _, loss = self.proj(x, y)
                outputs = (loss,) + outputs

        return outputs


@add_start_docstrings("""The XLM Model transformer with a language modeling head on top
    (linear layer with weights tied to the input embeddings). """,
    XLM_START_DOCSTRING, XLM_INPUTS_DOCSTRING)
class XLMWithLMHeadModel(XLMPreTrainedModel):
    r"""
        **labels**: (`optional`) ``torch.LongTensor`` of shape ``(batch_size, sequence_length)``:
            Labels for language modeling.
            Note that the labels **are shifted** inside the model, i.e. you can set ``lm_labels = input_ids``
            Indices are selected in ``[-1, 0, ..., config.vocab_size]``
            All labels set to ``-1`` are ignored (masked), the loss is only
            computed for labels in ``[0, ..., config.vocab_size]``

    Outputs: `Tuple` comprising various elements depending on the configuration (config) and inputs:
        **loss**: (`optional`, returned when ``labels`` is provided) ``torch.FloatTensor`` of shape ``(1,)``:
            Language modeling loss.
        **prediction_scores**: ``torch.FloatTensor`` of shape ``(batch_size, sequence_length, config.vocab_size)``
            Prediction scores of the language modeling head (scores for each vocabulary token before SoftMax).
        **hidden_states**: (`optional`, returned when ``config.output_hidden_states=True``)
            list of ``torch.FloatTensor`` (one for the output of each layer + the output of the embeddings)
            of shape ``(batch_size, sequence_length, hidden_size)``:
            Hidden-states of the model at the output of each layer plus the initial embedding outputs.
        **attentions**: (`optional`, returned when ``config.output_attentions=True``)
            list of ``torch.FloatTensor`` (one for each layer) of shape ``(batch_size, num_heads, sequence_length, sequence_length)``:
            Attentions weights after the attention softmax, used to compute the weighted average in the self-attention heads.

    Examples::

        tokenizer = XLMTokenizer.from_pretrained('xlm-mlm-en-2048')
        model = XLMWithLMHeadModel.from_pretrained('xlm-mlm-en-2048')
        input_ids = torch.tensor(tokenizer.encode("Hello, my dog is cute")).unsqueeze(0)  # Batch size 1
        outputs = model(input_ids)
        last_hidden_states = outputs[0]  # The last hidden-state is the first element of the output tuple

    """
    def __init__(self, config):
        super(XLMWithLMHeadModel, self).__init__(config)
        self.transformer = XLMModel(config)
        self.pred_layer = XLMPredLayer(config)

        self.init_weights()
        self.tie_weights()

    def tie_weights(self):
        """ Make sure we are sharing the embeddings
        """
        self._tie_or_clone_weights(self.pred_layer.proj, self.transformer.embeddings)

    def forward(self, input_ids, attention_mask=None, langs=None, token_type_ids=None, position_ids=None,
                lengths=None, cache=None, head_mask=None, labels=None):
        transformer_outputs = self.transformer(input_ids,
                                               attention_mask=attention_mask,
                                               langs=langs,
                                               token_type_ids=token_type_ids,
                                               position_ids=position_ids,
                                               lengths=lengths, 
                                               cache=cache,
                                               head_mask=head_mask)

        output = transformer_outputs[0]
        outputs = self.pred_layer(output, labels)
        outputs = outputs + transformer_outputs[1:]  # Keep new_mems and attention/hidden states if they are here

        return outputs


@add_start_docstrings("""XLM Model with a sequence classification/regression head on top (a linear layer on top of
    the pooled output) e.g. for GLUE tasks. """,
    XLM_START_DOCSTRING, XLM_INPUTS_DOCSTRING)
class XLMForSequenceClassification(XLMPreTrainedModel):
    r"""
        **labels**: (`optional`) ``torch.LongTensor`` of shape ``(batch_size,)``:
            Labels for computing the sequence classification/regression loss.
            Indices should be in ``[0, ..., config.num_labels - 1]``.
            If ``config.num_labels == 1`` a regression loss is computed (Mean-Square loss),
            If ``config.num_labels > 1`` a classification loss is computed (Cross-Entropy).

    Outputs: `Tuple` comprising various elements depending on the configuration (config) and inputs:
        **loss**: (`optional`, returned when ``labels`` is provided) ``torch.FloatTensor`` of shape ``(1,)``:
            Classification (or regression if config.num_labels==1) loss.
        **logits**: ``torch.FloatTensor`` of shape ``(batch_size, config.num_labels)``
            Classification (or regression if config.num_labels==1) scores (before SoftMax).
        **hidden_states**: (`optional`, returned when ``config.output_hidden_states=True``)
            list of ``torch.FloatTensor`` (one for the output of each layer + the output of the embeddings)
            of shape ``(batch_size, sequence_length, hidden_size)``:
            Hidden-states of the model at the output of each layer plus the initial embedding outputs.
        **attentions**: (`optional`, returned when ``config.output_attentions=True``)
            list of ``torch.FloatTensor`` (one for each layer) of shape ``(batch_size, num_heads, sequence_length, sequence_length)``:
            Attentions weights after the attention softmax, used to compute the weighted average in the self-attention heads.

    Examples::

        tokenizer = XLMTokenizer.from_pretrained('xlm-mlm-en-2048')
        model = XLMForSequenceClassification.from_pretrained('xlm-mlm-en-2048')
        input_ids = torch.tensor(tokenizer.encode("Hello, my dog is cute")).unsqueeze(0)  # Batch size 1
        labels = torch.tensor([1]).unsqueeze(0)  # Batch size 1
        outputs = model(input_ids, labels=labels)
        loss, logits = outputs[:2]

    """
    def __init__(self, config):
        super(XLMForSequenceClassification, self).__init__(config)
        self.num_labels = config.num_labels

        self.transformer = XLMModel(config)
        self.sequence_summary = SequenceSummary(config)

        self.init_weights()

    def forward(self, input_ids, attention_mask=None, langs=None, token_type_ids=None, position_ids=None,
                lengths=None, cache=None, head_mask=None, labels=None):
        transformer_outputs = self.transformer(input_ids,
                                               attention_mask=attention_mask,
                                               langs=langs,
                                               token_type_ids=token_type_ids,
                                               position_ids=position_ids,
                                               lengths=lengths, 
                                               cache=cache,
                                               head_mask=head_mask)

        output = transformer_outputs[0]
        logits = self.sequence_summary(output)

        outputs = (logits,) + transformer_outputs[1:]  # Keep new_mems and attention/hidden states if they are here

        if labels is not None:
            if self.num_labels == 1:
                #  We are doing regression
                loss_fct = MSELoss()
                loss = loss_fct(logits.view(-1), labels.view(-1))
            else:
                loss_fct = CrossEntropyLoss()
                loss = loss_fct(logits.view(-1, self.num_labels), labels.view(-1))
            outputs = (loss,) + outputs

        return outputs


@add_start_docstrings("""XLM Model with a span classification head on top for extractive question-answering tasks like SQuAD (a linear layers on top of
    the hidden-states output to compute `span start logits` and `span end logits`). """,
    XLM_START_DOCSTRING, XLM_INPUTS_DOCSTRING)
class XLMForQuestionAnsweringSimple(XLMPreTrainedModel):
    r"""
        **start_positions**: (`optional`) ``torch.LongTensor`` of shape ``(batch_size,)``:
            Labels for position (index) of the start of the labelled span for computing the token classification loss.
            Positions are clamped to the length of the sequence (`sequence_length`).
            Position outside of the sequence are not taken into account for computing the loss.
        **end_positions**: (`optional`) ``torch.LongTensor`` of shape ``(batch_size,)``:
            Labels for position (index) of the end of the labelled span for computing the token classification loss.
            Positions are clamped to the length of the sequence (`sequence_length`).
            Position outside of the sequence are not taken into account for computing the loss.
        **is_impossible**: (`optional`) ``torch.LongTensor`` of shape ``(batch_size,)``:
            Labels whether a question has an answer or no answer (SQuAD 2.0)
        **cls_index**: (`optional`) ``torch.LongTensor`` of shape ``(batch_size,)``:
            Labels for position (index) of the classification token to use as input for computing plausibility of the answer.
        **p_mask**: (`optional`) ``torch.LongTensor`` of shape ``(batch_size, sequence_length)``:
            Optional mask of tokens which can't be in answers (e.g. [CLS], [PAD], ...) 

    Outputs: `Tuple` comprising various elements depending on the configuration (config) and inputs:
        **loss**: (`optional`, returned when ``labels`` is provided) ``torch.FloatTensor`` of shape ``(1,)``:
            Total span extraction loss is the sum of a Cross-Entropy for the start and end positions.
        **start_scores**: ``torch.FloatTensor`` of shape ``(batch_size, sequence_length,)``
            Span-start scores (before SoftMax).
        **end_scores**: ``torch.FloatTensor`` of shape ``(batch_size, sequence_length,)``
            Span-end scores (before SoftMax).
        **hidden_states**: (`optional`, returned when ``config.output_hidden_states=True``)
            list of ``torch.FloatTensor`` (one for the output of each layer + the output of the embeddings)
            of shape ``(batch_size, sequence_length, hidden_size)``:
            Hidden-states of the model at the output of each layer plus the initial embedding outputs.
        **attentions**: (`optional`, returned when ``config.output_attentions=True``)
            list of ``torch.FloatTensor`` (one for each layer) of shape ``(batch_size, num_heads, sequence_length, sequence_length)``:
            Attentions weights after the attention softmax, used to compute the weighted average in the self-attention heads.

    Examples::

        tokenizer = XLMTokenizer.from_pretrained('xlm-mlm-en-2048')
        model = XLMForQuestionAnsweringSimple.from_pretrained('xlm-mlm-en-2048')
        input_ids = torch.tensor(tokenizer.encode("Hello, my dog is cute")).unsqueeze(0)  # Batch size 1
        start_positions = torch.tensor([1])
        end_positions = torch.tensor([3])
        outputs = model(input_ids, start_positions=start_positions, end_positions=end_positions)
        loss, start_scores, end_scores = outputs[:2]

    """
    def __init__(self, config):
        super(XLMForQuestionAnsweringSimple, self).__init__(config)

        self.transformer = XLMModel(config)
        self.qa_outputs = nn.Linear(config.hidden_size, config.num_labels)

        self.init_weights()

    def forward(self, input_ids, attention_mask=None, langs=None, token_type_ids=None, position_ids=None,
                lengths=None, cache=None, head_mask=None, start_positions=None, end_positions=None):
        transformer_outputs = self.transformer(input_ids,
                                               attention_mask=attention_mask,
                                               langs=langs,
                                               token_type_ids=token_type_ids,
                                               position_ids=position_ids,
                                               lengths=lengths, 
                                               cache=cache,
                                               head_mask=head_mask)

        sequence_output = transformer_outputs[0]

        logits = self.qa_outputs(sequence_output)
        start_logits, end_logits = logits.split(1, dim=-1)
        start_logits = start_logits.squeeze(-1)
        end_logits = end_logits.squeeze(-1)

        outputs = (start_logits, end_logits,)
        if start_positions is not None and end_positions is not None:
            # If we are on multi-GPU, split add a dimension
            if len(start_positions.size()) > 1:
                start_positions = start_positions.squeeze(-1)
            if len(end_positions.size()) > 1:
                end_positions = end_positions.squeeze(-1)
            # sometimes the start/end positions are outside our model inputs, we ignore these terms
            ignored_index = start_logits.size(1)
            start_positions.clamp_(0, ignored_index)
            end_positions.clamp_(0, ignored_index)

            loss_fct = CrossEntropyLoss(ignore_index=ignored_index)
            start_loss = loss_fct(start_logits, start_positions)
            end_loss = loss_fct(end_logits, end_positions)
            total_loss = (start_loss + end_loss) / 2
            outputs = (total_loss,) + outputs

        outputs = outputs + transformer_outputs[1:]  # Keep new_mems and attention/hidden states if they are here

        return outputs


@add_start_docstrings("""XLM Model with a beam-search span classification head on top for extractive question-answering tasks like SQuAD (a linear layers on top of
    the hidden-states output to compute `span start logits` and `span end logits`). """,
    XLM_START_DOCSTRING, XLM_INPUTS_DOCSTRING)
class XLMForQuestionAnswering(XLMPreTrainedModel):
    r"""
        **start_positions**: (`optional`) ``torch.LongTensor`` of shape ``(batch_size,)``:
            Labels for position (index) of the start of the labelled span for computing the token classification loss.
            Positions are clamped to the length of the sequence (`sequence_length`).
            Position outside of the sequence are not taken into account for computing the loss.
        **end_positions**: (`optional`) ``torch.LongTensor`` of shape ``(batch_size,)``:
            Labels for position (index) of the end of the labelled span for computing the token classification loss.
            Positions are clamped to the length of the sequence (`sequence_length`).
            Position outside of the sequence are not taken into account for computing the loss.
        **is_impossible**: (`optional`) ``torch.LongTensor`` of shape ``(batch_size,)``:
            Labels whether a question has an answer or no answer (SQuAD 2.0)
        **cls_index**: (`optional`) ``torch.LongTensor`` of shape ``(batch_size,)``:
            Labels for position (index) of the classification token to use as input for computing plausibility of the answer.
        **p_mask**: (`optional`) ``torch.LongTensor`` of shape ``(batch_size, sequence_length)``:
            Optional mask of tokens which can't be in answers (e.g. [CLS], [PAD], ...) 

    Outputs: `Tuple` comprising various elements depending on the configuration (config) and inputs:
        **loss**: (`optional`, returned when ``labels`` is provided) ``torch.FloatTensor`` of shape ``(1,)``:
            Total span extraction loss is the sum of a Cross-Entropy for the start and end positions.
        **start_scores**: ``torch.FloatTensor`` of shape ``(batch_size, sequence_length,)``
            Span-start scores (before SoftMax).
        **end_scores**: ``torch.FloatTensor`` of shape ``(batch_size, sequence_length,)``
            Span-end scores (before SoftMax).
        **hidden_states**: (`optional`, returned when ``config.output_hidden_states=True``)
            list of ``torch.FloatTensor`` (one for the output of each layer + the output of the embeddings)
            of shape ``(batch_size, sequence_length, hidden_size)``:
            Hidden-states of the model at the output of each layer plus the initial embedding outputs.
        **attentions**: (`optional`, returned when ``config.output_attentions=True``)
            list of ``torch.FloatTensor`` (one for each layer) of shape ``(batch_size, num_heads, sequence_length, sequence_length)``:
            Attentions weights after the attention softmax, used to compute the weighted average in the self-attention heads.

    Examples::

        tokenizer = XLMTokenizer.from_pretrained('xlm-mlm-en-2048')
        model = XLMForQuestionAnswering.from_pretrained('xlm-mlm-en-2048')
        input_ids = torch.tensor(tokenizer.encode("Hello, my dog is cute")).unsqueeze(0)  # Batch size 1
        start_positions = torch.tensor([1])
        end_positions = torch.tensor([3])
        outputs = model(input_ids, start_positions=start_positions, end_positions=end_positions)
        loss, start_scores, end_scores = outputs[:2]

    """
    def __init__(self, config):
        super(XLMForQuestionAnswering, self).__init__(config)

        self.transformer = XLMModel(config)
        self.qa_outputs = SQuADHead(config)

        self.init_weights()

    def forward(self, input_ids, attention_mask=None, langs=None, token_type_ids=None, position_ids=None,
                lengths=None, cache=None, head_mask=None, start_positions=None, end_positions=None,
                is_impossible=None, cls_index=None, p_mask=None):
        transformer_outputs = self.transformer(input_ids,
                                               attention_mask=attention_mask,
                                               langs=langs,
                                               token_type_ids=token_type_ids,
                                               position_ids=position_ids,
                                               lengths=lengths, 
                                               cache=cache,
                                               head_mask=head_mask)

        output = transformer_outputs[0]

        outputs = self.qa_outputs(output, start_positions=start_positions, end_positions=end_positions,
                                  cls_index=cls_index, is_impossible=is_impossible, p_mask=p_mask)

        outputs = outputs + transformer_outputs[1:]  # Keep new_mems and attention/hidden states if they are here

        return outputs



import abc
from relogic.logickit.scorer.scorer import Scorer
from relogic.logickit.scorer.sent_level_scorer import SentLevelScorer
from relogic.logickit.utils.utils import softmax
import json
import os
import numpy as np


class F1Score(SentLevelScorer, metaclass=abc.ABCMeta):
  def __init__(self):
    super(F1Score, self).__init__()
    self._n_correct, self._n_predicted, self._n_gold = 0, 0, 0

  def _get_results(self):
    if self._n_correct == 0:
      p, r, f1 = 0, 0, 0
    else:
      p = 100.0 * self._n_correct / self._n_predicted
      r = 100.0 * self._n_correct / self._n_gold
      f1 = 2 * p * r / (p + r)
    return [
      ("total label", self._n_gold),
      ("find label", self._n_predicted),
      ("correct label", self._n_correct),
      ("precision", p),
      ("recall", r),
      ("f1", f1),
      ("loss", self.get_loss()),
    ]

class MultiClassAccuracyScorer(Scorer):
  def __init__(self, label_mapping, dump_to_file, dataflow):
    super(MultiClassAccuracyScorer, self).__init__()
    self._n_correct, self._n_total = 0, 0
    self.dataflow = dataflow
    self._examples = []
    self._preds = []
    if dump_to_file:
      self.dump_to_file_path = os.path.join(dump_to_file["output_dir"], dump_to_file["task_name"] + "_dump.json")
      self.dump_to_file_handler = open(self.dump_to_file_path, 'w')

  def update(self, mbs, predictions, loss, extra_args):
    super(MultiClassAccuracyScorer, self).update(mbs, predictions, loss, extra_args)
    labels = self.dataflow.decode_to_labels(preds=predictions, mb=mbs)
    self._examples.extend(mbs.examples)
    self._preds.extend(labels)

  def _get_results(self):
    for example, pred in zip(self._examples, self._preds):
      if set(example.labels) == set(pred):
        self._n_correct += 1
      self._n_total += 1
    return [("accuracy", self._n_correct / self._n_total)]

  def get_loss(self):
    return 0




class RelationF1Scorer(F1Score):
  def __init__(self, label_mapping, dump_to_file):
    super(RelationF1Scorer, self).__init__()
    self._inv_label_mapping = {v: k for k, v in label_mapping.items()}
    self._o = 'no_relation'
    self.counter = 0
    if dump_to_file:
      self.dump_to_file_path = os.path.join(dump_to_file["output_dir"], dump_to_file["task_name"] + "_dump.json")
      self.dump_to_file_handler = open(self.dump_to_file_path, 'w')
      attention_map_path = os.path.join(dump_to_file["output_dir"], "attn_maps")
      if not os.path.exists(attention_map_path):
        os.mkdir(attention_map_path)
      self.attention_dump_file_path = os.path.join(attention_map_path, dump_to_file["task_name"] + "_attention")
      # self.attention_dump_file_handler = open(self.attention_dump_file_path, 'wb')
    self._attn_maps = []
    self.counter = 0


  def update(self, mbs, predictions, loss, extra_args):
    super(RelationF1Scorer, self).update(mbs, predictions, loss, extra_args)
    if self.need_to_clear_output:
      self.dump_to_file_handler = open(self.dump_to_file_path, 'w')
      self.need_to_clear_output = False
    n_sents = 0

    for idx, (example, preds) in enumerate(zip(mbs.examples, predictions)):
      self._examples.append(example)
      self._preds.append(preds)
      pred_tag = preds.argmax(-1).item()
      if self.dump_to_file_path:
        self.dump_to_file_handler.write(
          json.dumps({
            "text": example.raw_text,
            "token": example.tokens,
            "subject": example.subj_text,
            "object": example.obj_text,
            "label": example.label,
            "predicted": self._inv_label_mapping[pred_tag]
          }) + "\n")
        if "attention_map" in extra_args:
          np.save(self.attention_dump_file_path + str(self.counter), extra_args["attention_map"][idx])
          self.counter += 1
      # do not dumping heavy results on memory
      n_sents += 1

    self._total_loss += loss * n_sents
    self._total_sents += n_sents

  def _get_results(self):
    self.need_to_clear_output = True

    self._n_correct, self._n_predicted, self._n_gold = 0, 0, 0
    # data_to_dump = []

    for example, preds in zip(self._examples, self._preds):
      pred_tag = preds.argmax(-1).item()

      confidence = max(softmax(preds.data.cpu().numpy()))
      if self._inv_label_mapping[pred_tag] == example.label and \
        example.label != self._o:
          self._n_correct += 1
      if example.label != self._o:
        self._n_gold += 1
      if self._inv_label_mapping[pred_tag] != self._o:
        self._n_predicted += 1
      # if self.dump_to_file_path:
      #   data_to_dump.append((int(example.guid), example.raw_text,
      #                        example.subj_text, example.obj_text, example.label,
      #                        self._inv_label_mapping[pred_tag], confidence))

    # if self.dump_to_file_path:
    #   data_to_dump = sorted(data_to_dump, key=lambda x: x[0])
    #   for idx, raw_text, subj, obj, gold_label, pred_label, confidence in data_to_dump:
    #     json_to_write = {
    #       "text": raw_text,
    #       "subject": subj,
    #       "object": obj,
    #       "relation": gold_label,
    #       "predicted_relation": pred_label,
    #       "confidence": confidence
    #     }
    #     self.dump_to_file_handler.write(json.dumps(json_to_write) + "\n")
    #
    #
    # if self.dump_to_file_path:
    #   self.dump_to_file_handler.close()


    return super(RelationF1Scorer, self)._get_results()


import sys
import json
from typing import List, Dict
from relogic.logickit.scripts.squad20_eval import normalize_answer, compute_exact, compute_f1
from itertools import groupby
import argparse
import operator

def calculate_score(gold_answer_list: List, pred_answer_list: List):
  if len(gold_answer_list) == 0:
    if len(pred_answer_list) == 0:
      return (1.0, 1.0, 1.0)
    else:
      return (0.0, 1.0, 0.0)
  elif len(pred_answer_list) == 0:
    return (1.0, 0.0, 0.0)
  else:
    pred_answer_list = list(set([normalize_answer(ans) for ans in pred_answer_list]))
    gold_answer_list = list(set([normalize_answer(ans) for ans in gold_answer_list]))

    c = 0
    for ans in pred_answer_list:
      if ans in gold_answer_list:
        c += 1
    p = c / len(pred_answer_list)
    r = c / len(gold_answer_list)
    if p == 0 or r == 0:
      f1 = 0
    else:
      f1 = 2 * p * r / (p+r)
    return (p, r, f1)

def calculate_score_max_over_ground_truths(gold_answer_list: List, pred_answer_list: List):
  """Given a prediction and multiple valid answers, return the score of
  the best prediction-answer_n pair based on exact match"""
  best_scores_for_ground_truths = (0, 0)

  for ground_truth in gold_answer_list:
    for predicted_ans in pred_answer_list:
      em = compute_exact(ground_truth, predicted_ans)
      f1 = compute_f1(ground_truth, predicted_ans)
      if best_scores_for_ground_truths[0] < em:
        best_scores_for_ground_truths = (em, f1)
  return best_scores_for_ground_truths

def rc_evalutate(gold_data: List, pred_answers: List):
  if len(gold_data) != len(pred_answers):
    print("Gold data length: {} != Pred answers length: {}".format(
      len(gold_data), len(pred_answers)))
  total, f1_sum, em_sum = 0, 0, 0
  dump_results = []
  for entry, prediction in zip(gold_data, pred_answers):
    if (len(entry["Parses"]) == 0):
      print("Enpty Parses in gold set. Exist!")
      exit()
    total += 1

    qid, prediction = prediction[0], prediction[1] # (id, list of answers)
    all_parse_answers = []
    for parse in entry["Parses"]:
      parse_answers = [ans["EntityName"] if ans["EntityName"] is not None else ans["AnswerArgument"]
                       for ans in parse["Answers"]]
      all_parse_answers.extend(parse_answers)
    all_parse_answers = list(set(all_parse_answers))
    em, f1 = calculate_score_max_over_ground_truths(all_parse_answers, prediction)
    dump_results.append((qid, prediction, f1, em, all_parse_answers))
    f1_sum += f1
    em_sum += em

  return [("f1", f1_sum / total),
          ("exact_match", em_sum / total)], dump_results

def sp_evaluate(gold_data: List, pred_answers: List):
  if len(gold_data) != len(pred_answers):
    print("Gold data length: {} != Pred answers length: {}".format(
      len(gold_data), len(pred_answers)))
  total, f1_sum, rec_sum, prec_sum, num_correct = 0, 0, 0, 0, 0

  for entry, prediction in zip(gold_data, pred_answers):
    if (len(entry["Parses"]) == 0):
      print("Enpty Parses in gold set. Exist!")
      exit()
    total += 1

    best_f1 = -1
    best_rec = -1
    best_prec = -1
    prediction = prediction[1] # (id, list of answers)

    for parse in entry["Parses"]:
      parse_answers = [ans["EntityName"] if ans["EntityName"] is not None else ans["AnswerArgument"]
                       for ans in parse["Answers"]]


      prec, rec, f1 = calculate_score(parse_answers, prediction)
      if f1 > best_f1:
        best_f1 = f1
        best_rec = rec
        best_prec = prec

    f1_sum += best_f1
    rec_sum += best_rec
    prec_sum += best_prec

    if best_f1 == 1.0:
      num_correct += 1
  precision = prec_sum / total
  recall = rec_sum / total
  if precision == 0 or recall == 0:
    f1 = 0
  else:
    f1 = 2 * precision * recall / (precision + recall)
  return [
    ("f1", f1),
    ("P", precision),
    ("R", recall),
    ("exact_match", num_correct / total)]

def aggregate_answers(pred_answers: Dict, span_scores: Dict) -> List:
  groups = groupby(pred_answers.items(), key=lambda x: x[0].split('.')[0])
  prediction = []
  for k, vs in groups:
    # We use confidence score based voting method to aggregate results
    answer_text_score = {}
    for v in vs:
      if len(v[1]) > 0:
        # Filter out empty span
        if v[1] not in answer_text_score:
          answer_text_score[v[1]] = 0
        answer_text_score[v[1]] += span_scores[v[0]]
    if len(answer_text_score) == 0:
      best_ans = ""
    else:
      best_ans = max(answer_text_score, key=answer_text_score.get)
    prediction.append((k, [best_ans]))
  prediction = sorted(prediction, key=lambda x: int(x[0]), reverse=False)
  return prediction

if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("--gold_file", type=str)
  parser.add_argument("--pred_file", type=str)
  parser.add_argument("--ir_file", type=str)
  parser.add_argument("--span_score_file", type=str)
  parser.add_argument("--dump_file", type=str)

  args = parser.parse_args()

  gold_data = json.load(open(args.gold_file))["Questions"]

  pred_answers = json.load(open(args.pred_file))

  span_score_file = json.load(open(args.span_score_file))
  ir_score_file = None

  pred_answers = aggregate_answers(pred_answers, span_score_file)

  metrics, dump_results = rc_evalutate(gold_data, pred_answers)
  print(metrics)

  with open(args.dump_file, "w") as fout:
    json.dump(dump_results, fout, indent=2)




import argparse
import json
import copy

from relogic.logickit.utils import get_span_labels

def ner_analysis(file_path):
  """
  We assume the file is in the following format
  {"tokens": List[str],
   "labels": List[str],
   "predicted_labels": List[str]}

  We categorize the error types into four types:
    type-error: correct entity boundary but wrong entity type
    boundary-error: wrong boundary
    extra-entity-error: predicting non-entity phrase as extra entity
    miss-entity-error: missing an entity
  """
  examples = []
  with open(file_path) as fin:
    for line in fin:
      examples.append(json.loads(line))
  type_error = 0
  boundary_error = 0
  extra_entity_error = 0
  miss_entity_error = 0
  for example in examples:
    gold_entities, _ = get_span_labels(example["labels"])
    pred_entities, _ = get_span_labels(example["predicted_labels"])
    if len(gold_entities - pred_entities) == 0:
      continue
    else:
      error_summary = {}
      error_summary["tokens"] = " ".join(example["tokens"])
      error_summary["type-error"] = []
      error_summary["boundary-error"] = []
      error_summary["extra-entity-error"] = []
      error_summary["miss-entity-error"] = []
      potential_miss_entity = gold_entities - pred_entities
      potential_extra_entity = pred_entities - gold_entities
      copy_potential_extra_entity = copy.deepcopy(potential_extra_entity)
      for span in potential_extra_entity:
        for gold_entity in gold_entities:
          if span[0] == gold_entity[0] and span[1] == gold_entity[1]:
            # This is type-error
            if span in copy_potential_extra_entity:
              error_summary["type-error"].append({
                "gold_entity_str": " ".join(example["tokens"][gold_entity[0]: gold_entity[1]+1]),
                "gold_entity_span": gold_entity,
                "pred_entity_str": " ".join(example["tokens"][span[0]: span[1]+1]),
                "pred_entity_span": span})
              type_error += 1
              copy_potential_extra_entity.remove(span)
              if gold_entity in potential_miss_entity:
                potential_miss_entity.remove(gold_entity)
          elif (span[0] >= gold_entity[0] and span[0] <= gold_entity[1]) \
                or (span[1] >= gold_entity[0] and span[1] <= gold_entity[1]) \
                or (span[0] <= gold_entity[0] and span[1] >= gold_entity[1]) \
                or (span[0] >= gold_entity[0] and span[1] <= gold_entity[1]):
            if span in copy_potential_extra_entity:
              error_summary["boundary-error"].append({
                "gold_entity_str": " ".join(example["tokens"][gold_entity[0]: gold_entity[1]+1]),
                "gold_entity_span": gold_entity,
                "pred_entity_str": " ".join(example["tokens"][span[0]: span[1]+1]),
                "pred_entity_span": span})
              boundary_error += 1
              copy_potential_extra_entity.remove(span)
              if gold_entity in potential_miss_entity:
                potential_miss_entity.remove(gold_entity)
      for span in copy_potential_extra_entity:
        error_summary["extra-entity-error"].append({
            "pred_entity_str": " ".join(example["tokens"][span[0]: span[1] + 1]),
            "pred_entity_span": span})
        extra_entity_error += 1
      for span in potential_miss_entity:
        error_summary["miss-entity-error"].append({
              "gold_entity_str": " ".join(example["tokens"][span[0]: span[1]+1]),
              "gold_entity_span": span})
        miss_entity_error += 1
    print(json.dumps(error_summary, indent=2))
  total_error_count = type_error + boundary_error + extra_entity_error + miss_entity_error
  print("type-error", type_error, type_error / total_error_count)
  print("boundary-error", boundary_error, boundary_error / total_error_count)
  print("extra-entity-error", extra_entity_error, extra_entity_error / total_error_count)
  print("miss-entity-error", miss_entity_error, miss_entity_error / total_error_count)


if __name__=="__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("--task_name", choices=["ner"])
  parser.add_argument("--dump_file_path", type=str)

  args = parser.parse_args()

  if args.task_name == "ner":
    ner_analysis(args.dump_file_path)


"""
The module contains the implementation of DataFlow on SRL task.
"""

import json
from typing import List, Tuple
import os

import torch

from relogic.logickit.base.constants import SRL_LABEL_SEQ_BASED, SRL_LABEL_SPAN_BASED, SRL_PREDICATE_EXTRA_SURFACE
from relogic.logickit.dataflow import DataFlow, Example, Feature, MiniBatch
from relogic.logickit.tokenizer.fasttext_tokenization import FasttextTokenizer
from relogic.logickit.utils import create_tensor, get_span_labels
from relogic.structures import enumerate_spans
from transformers.tokenization_utils import PreTrainedTokenizer

MAX_SEQ_LENGTH=500
MAX_VERB_LENGTH=9

class SRLExample(Example):
  """SRLExample contains the attributes and functionality of an SRL example.

  Args:
    text (str): A sentence string.
    labels (List[Tuple] or List[Str]): A list of labels. Each label is a Tuple containing
      predicate_start_index, predicate_end_index, predicate_text_string,
      argument_start_index, argument_end_index, argument_text_string, argument_label.

      If it is sequence labeling, then the labels are a sequence of labels string.

  """
  def __init__(self,
               text,
               predicate_text=None,
               predicate_index=None,
               labels=None,
               pos_tag_label=None,
               predicate_sense_label=None,
               label_candidates=None):
    super(SRLExample, self).__init__()
    self.text = text
    self.raw_tokens = text.split(" ")
    self.labels = labels
    # Hard code here
    self.label_padding = 'X'
    self.predicate_text = predicate_text
    if self.predicate_text is not None:
      self.raw_predicate_text_tokens = self.predicate_text.split()
    self.predicate_index = predicate_index
    # We assume the index has only one value

    self.pos_tag_label = pos_tag_label
    self.predicate_sense_label = predicate_sense_label
    self.label_candidates = label_candidates

    self.is_valid = True
    # I add this tag for long sentence. If the sentence is long, the tokenizer
    # will cut it off. But if the predicate index is large and the corresponding
    # position is not in sentence anymore, the example becomes an invalid example.
    # The return label will be empty.

  def process(self, tokenizers, *inputs, **kwargs):
    """Process the SRL example.

    This process requires the tokenizer. Furthermore, if this example is for
     training and evaluation, the label_format and label_mapping are required.

    History Log: Originally we use adapted tokenizer for determining the token head
      for the pretokenized sentence. Here we plan to change it with for loop.
      Then we do not need to adapt it for each tokenizer.
      ```
      self.text_tokens, self.text_is_head = tokenizer.tokenize(self.text)
      ```
      We will adapt the code from transformer example run_ner.py and utils_ner.py
    """
    use_gold_predicate = kwargs.pop("use_gold_predicate", False)
    use_gold_argument = kwargs.pop("use_gold_argument", False)
    predicate_reveal_method = kwargs.pop("predicate_reveal_method", SRL_PREDICATE_EXTRA_SURFACE)
    for tokenizer in tokenizers.values():
      if isinstance(tokenizer, PreTrainedTokenizer):
        # BERT process part
        self.text_tokens = []
        self.text_is_head = []
        for word in self.raw_tokens:
          word_tokens = tokenizer.tokenize(word)
          if len(word_tokens) == 0:
            self.text_tokens.append("[UNK]")
            self.text_is_head.append(1)
          else:
            self.text_tokens.extend(word_tokens)
            self.text_is_head.extend([1] + [0] * (len(word_tokens) - 1))
        if len(self.text_tokens) > MAX_SEQ_LENGTH:
          self.text_tokens = self.text_tokens[:MAX_SEQ_LENGTH]
          self.text_is_head = self.text_is_head[:MAX_SEQ_LENGTH]

        # Currently it only support one group of model (BERT).
        self.tokens = [tokenizer.cls_token] + self.text_tokens + [tokenizer.sep_token]
        self.segment_ids = [0] * (len(self.text_tokens) + 2)
        self.is_head = [2] + self.text_is_head + [2]
        self.head_index = [
            idx for idx, value in enumerate(self.is_head) if value == 1
        ] + [len(self.is_head) - 1]
        # index of [SEP] is len(self.is_head) - 1.
        # Assume we have sentence [A, B, C]. After processing,
        # it becomes [[CLS], A, B, C, [SEP]]
        # length of head_index is 3, that is [1, 2, 3].
        # We pad it with len(self.is_head) - 1, which is 4.
        # Assume we have span (2, 3) = C, exclusive.
        # self.head_index[2] =  3, self.head_index[3] = 4.
        # So the span for tokenized sentence is (3, 4) = C
        assert (len(self.tokens) == len(self.is_head))

        # This part is mainly for the arxiv version paper.
        if predicate_reveal_method == SRL_PREDICATE_EXTRA_SURFACE:
          # In this case we assume the input only has one predicate
          # User needs to separate them by themselves.

          assert self.predicate_text is not None
          predicate_text_tokens = []
          for word in self.raw_predicate_text_tokens:
            word_tokens = tokenizer.tokenize(word)
            predicate_text_tokens.extend(word_tokens)
            # self.text_is_head.extend([1] + [0] * (len(word_tokens) - 1))
            # We ignore the is_head information for predicate for now
          predicate_text_tokens = predicate_text_tokens[:MAX_VERB_LENGTH]
          # Limit the verb length based on the MAX SEQ LENGTH

          predicate_text_tokens += [tokenizer.sep_token]
          self.tokens.extend(predicate_text_tokens)
          self.segment_ids += [1] * len(predicate_text_tokens)
          assert self.predicate_index is not None
          if self.predicate_index >= len(self.head_index)-1:
            self.is_valid = False
            self.predicate_idx = self.head_index[-2]
            # Just put a random index
          else:
            self.predicate_idx = self.head_index[self.predicate_index]


        self.input_ids = tokenizer.convert_tokens_to_ids(self.tokens)
        self.input_mask = [1] * len(self.input_ids)

        # Enumerate the span and map the spans into sub-token level
        if use_gold_argument:
          assert self.labels is not None
          spans = []
          for label in self.labels:
            (predicate_start, predicate_end, predicate_text,
             arg_start, arg_end, argument_text, arg_label) = label
            span_tuple = (arg_start, arg_end)
            if span_tuple not in spans:
              spans.append(span_tuple)
        else:
          spans = enumerate_spans(sentence=self.raw_tokens[:len(self.head_index)-1], max_span_width=30)

        self.enumerated_span_candidates: List[Tuple[int, int]] = []
        for span in spans:
          self.enumerated_span_candidates.append(
              (self.head_index[span[0]], self.head_index[span[1]]))

        # Currently we assume that the predicate length is 1
        if use_gold_predicate:
          assert self.labels is not None
          spans = []
          for label in self.labels:
            (predicate_start, predicate_end, predicate_text,
             arg_start, arg_end, argument_text, arg_label) = label
            span_tuple = (predicate_start, predicate_end)
            if span_tuple not in spans:
              spans.append(span_tuple)
        else:
          spans = enumerate_spans(sentence=self.raw_tokens[:len(self.head_index)-1], max_span_width=1)

        self.predicate_candidates: List[Tuple[int, int]] = []
        for span in spans:
          self.predicate_candidates.append(
              (self.head_index[span[0]], self.head_index[span[1]]))

        # If this is for model development, then self.labels is not None
        # If this is for deployment, then self.labels is None
        if self.labels is not None:
          label_mapping = kwargs.get("label_mapping")
          self.label_padding_id = label_mapping[self.label_padding]
          label_format = kwargs.get("label_format")

          if label_format == SRL_LABEL_SPAN_BASED:
            self.label_ids = []
            self.pred_span_label_dict = dict([(item, 0)
                                              for item in self.predicate_candidates])
            self.arg_span_label_dict = dict([(item, 0)
                                             for item in self.enumerated_span_candidates])

            for label in self.labels:
              (predicate_start, predicate_end, predicate_text,
               arg_start, arg_end, argument_text, arg_label) = label

              label_tuple = (self.head_index[predicate_start],
                             self.head_index[predicate_end],
                             self.head_index[arg_start],
                             self.head_index[arg_end], label_mapping[arg_label])
              self.label_ids.append(label_tuple)
              self.pred_span_label_dict[(self.head_index[predicate_start], self.head_index[predicate_end])] = 1
              self.arg_span_label_dict[(self.head_index[arg_start], self.head_index[arg_end])] = 1

            self.pred_span_label_ids = [self.pred_span_label_dict[item] for item in self.predicate_candidates]
            self.arg_span_label_ids = [self.arg_span_label_dict[item] for item in self.enumerated_span_candidates]

          elif label_format == SRL_LABEL_SEQ_BASED:
            label_mapping = kwargs.get("label_mapping")
            self.label_padding_id = label_mapping[self.label_padding]
            self.label_ids = [self.label_padding_id] * len(self.tokens)
            assert len(self.labels) == len(self.head_index[:-1])
            for idx, label in zip(self.head_index[:-1], self.labels):
              # we ignore the index of [SEP]
              self.label_ids[idx] = label_mapping[label]
          else:
            raise ValueError()

        # Process the pos tag labels into sequence labeling task
        if self.pos_tag_label is not None:
          pos_tag_label_mapping = kwargs.get("pos_tag_label_mapping")
          self.pos_tag_label_padding_id = pos_tag_label_mapping[self.label_padding]
          self.pos_tag_label_ids = [self.label_padding_id] * len(self.input_ids)
          for idx, label in zip(self.head_index, self.pos_tag_label):
            self.pos_tag_label_ids[idx] = pos_tag_label_mapping[label]

        if self.predicate_sense_label is not None:
          predicate_sense_mapping = kwargs.get("predicate_sense_mapping")
          self.predicate_sense_label_padding_id = predicate_sense_mapping[self.label_padding]
          self.predicate_sense_label_ids = [self.predicate_sense_label_padding_id] * len(self.input_ids)


      elif isinstance(tokenizer, FasttextTokenizer):
        # Traditional process part
        self._text_tokens = tokenizer.tokenize(self.text)
        # self._text_tokens, self._char_tokens = tokenizer.tokenize(self.text)
        self._input_token_ids = tokenizer.convert_tokens_to_ids(self._text_tokens)
        # self._input_char_ids = tokenizer.convert_char_to_ids(self._char_tokens)

        spans = enumerate_spans(sentence=self._text_tokens, max_span_width=20)

        self._enumerated_span_candidates: List[Tuple[int, int]] = []
        for span in spans:
          self._enumerated_span_candidates.append(span)

        spans = enumerate_spans(sentence=self._text_tokens, max_span_width=1)
        self._predicate_candidates: List[Tuple[int, int]] = []
        for span in spans:
          self._predicate_candidates.append(span)

        if self.labels is not None:
          label_mapping = kwargs.get("label_mapping")
          self.label_padding_id = label_mapping[self.label_padding]
          label_format = kwargs.get("label_format")

          if label_format == SRL_LABEL_SPAN_BASED:
            self._label_ids = []
            self._pred_span_label_dict = dict([(item, 0)
                                               for item in self._predicate_candidates])
            self._arg_span_label_dict = dict([(item, 0)
                                              for item in self._enumerated_span_candidates])
            for label in self.labels:
              (predicate_start, predicate_end, predicate_text,
               arg_start, arg_end, argument_text, arg_label) = label

              label_tuple = (predicate_start, predicate_end,
                             arg_start, arg_end, label_mapping[arg_label])
              self._label_ids.append(label_tuple)
              self._pred_span_label_dict[(predicate_start, predicate_end)] = 1
              self._arg_span_label_dict[(arg_start, arg_end)] = 1

            self._pred_span_label_ids = [self._pred_span_label_dict[item] for item in self._predicate_candidates]
            self._arg_span_label_ids = [self._arg_span_label_dict[item] for item in self._enumerated_span_candidates]
          elif label_format == SRL_LABEL_SEQ_BASED:
            pass
          else:
            raise ValueError

  @classmethod
  def from_structure(cls, structure):
    """Implementation of converting structure into SRLExample."""

    return cls(text=structure.tokenized_text,
               predicate_text=structure.predicate_text,
               predicate_index=structure.predicate_index)

  @classmethod
  def from_json(cls, example):
    """Implementation of converting json object into SRLExample.

    This object can be grouped all predicates together or represents a single predicate.

    """

    return cls(text=" ".join(example["tokens"]),
               labels=example.get("labels", None),
               pos_tag_label=example.get("pos_tag", None),
               predicate_sense_label=example.get("predicate_sense", None),
               label_candidates=example.get("label_candidates", None),
               predicate_text=example.get("predicate_text", None),
               predicate_index=example.get("predicate_index", None))

  @property
  def len(self):
    """SRL example tokens length (after all preprocessing)."""

    return len(self.tokens)

  @property
  def _len(self):
    """SRL example classical token length"""

    return len(self._input_token_ids)

  @property
  def label_len(self):
    """SRL example label length (after all preprocessing). Can be span-based or sequence-based"""

    if hasattr(self, "label_ids"):
      return len(self.label_ids)
    return 0

  @property
  def _label_len(self):
    """SRL example classical label length."""
    if hasattr(self, "_label_ids"):
      return len(self._label_ids)
    return 0


class SRLFeature(Feature):
  """SRL features.
  Some extra features such as arg_candidates and predicate_candidates

  """
  def __init__(self, *inputs, **kwargs):
    super(SRLFeature, self).__init__()
    # Bert based feature
    self.input_ids = kwargs.pop("input_ids")
    self.input_mask = kwargs.pop("input_mask")
    self.segment_ids = kwargs.pop("segment_ids")
    self.is_head = kwargs.pop("is_head")
    self.label_ids = kwargs.pop("label_ids")
    self.arg_candidates = kwargs.pop("arg_candidates")
    self.predicate_candidates = kwargs.pop("predicate_candidates")
    self.arg_candidate_label_ids = kwargs.pop("arg_candidate_label_ids")
    self.predicate_candidate_label_ids = kwargs.pop("predicate_candidate_label_ids")
    self.predicate_idx = kwargs.pop("predicate_idx")

    # Label embeddings
    self.label_candidates = kwargs.pop("label_candidates")
    self.label_candidates_mask = kwargs.pop("label_candidates_mask")

    # POS tag labels
    self.pos_tag_label_ids = kwargs.pop("pos_tag_label_ids")

    # Classical feature
    self._input_token_ids = kwargs.pop("_input_token_ids")
    self._label_ids = kwargs.pop("_label_ids")
    self._arg_candidates = kwargs.pop("_arg_candidates")
    self._predicate_candidates = kwargs.pop("_predicate_candidates")


class SRLMiniBatch(MiniBatch):
  """MiniBatch Implementation based on SRL task.

  """
  def __init__(self, *inputs, **kwargs):
    super(SRLMiniBatch, self).__init__(*inputs, **kwargs)

  def generate_input(self, device, use_label):
    """Generate tensors based on SRL Features."""
    # BERT based features
    inputs = {}
    inputs["task_name"] = self.task_name
    inputs["input_ids"] = create_tensor(self.input_features, "input_ids",
                                        torch.long, device)
    inputs["input_mask"] = create_tensor(self.input_features, "input_mask",
                                         torch.long, device)
    inputs["segment_ids"] = create_tensor(self.input_features, "segment_ids",
                                          torch.long, device)
    inputs["input_head"] = create_tensor(self.input_features, "is_head",
                                         torch.long, device)
    inputs["predicate_idx"] = create_tensor(self.input_features, "predicate_idx",
                                            torch.long, device)
    if use_label:
      label_ids = create_tensor(self.input_features, "label_ids",
                                          torch.long, device)
      predicate_candidate_label_ids = create_tensor(self.input_features, "predicate_candidate_label_ids",
                                           torch.long, device)
      arg_candidate_label_ids = create_tensor(self.input_features, "arg_candidate_label_ids",
                                           torch.long, device)
      pos_tag_label_ids = create_tensor(self.input_features, "pos_tag_label_ids", torch.long, device)
      if predicate_candidate_label_ids is None or arg_candidate_label_ids is None:
        inputs["label_ids"] = label_ids
      else:
        inputs["label_ids"] = (label_ids, predicate_candidate_label_ids, arg_candidate_label_ids, pos_tag_label_ids)
    else:
      inputs["label_ids"] = None
    inputs["arg_candidates"] = create_tensor(
        self.input_features, "arg_candidates", torch.long, device)
    inputs["predicate_candidates"] = create_tensor(
        self.input_features, "predicate_candidates", torch.long, device)

    # Label Embeddings
    inputs["label_candidates"] = create_tensor(self.input_features, "label_candidates", 
        torch.long, device)
    inputs["label_candidates_mask"] = create_tensor(self.input_features, "label_candidates_mask", 
        torch.long, device)
    # Classical features
    # inputs["_input_token_ids"] = create_tensor(self.input_features, "_input_token_ids",
    #                                            torch.long, device)
    # inputs["_token_length"] = create_tensor(self.input_features, "_token_length",
    #                                         torch.long, device)
    # if use_label:
    #   _label_ids = create_tensor(self.input_features, "_label_ids",
    #                                       torch.long, device)
    #   _predicate_candidate_label_ids = create_tensor(self.input_features, "_predicate_candidate_label_ids",
    #                                        torch.long, device)
    #   _arg_candidate_label_ids = create_tensor(self.input_features, "_arg_candidate_label_ids",
    #                                        torch.long, device)
    #   if _predicate_candidate_label_ids is None or _arg_candidate_label_ids is None:
    #     inputs["_label_ids"] = _label_ids
    #   else:
    #     inputs["_label_ids"] = (_label_ids, _predicate_candidate_label_ids, _arg_candidate_label_ids)
    # else:
    #   inputs["_label_ids"] = None
    # inputs["_arg_candidates"] = create_tensor(
    #     self.input_features, "_arg_candidates", torch.long, device)
    # inputs["_predicate_candidates"] = create_tensor(
    #   self.input_features, "_predicate_candidates", torch.long, device)
    # inputs["extra_args"] = {
    #   "selected_non_final_layers": [1, 9, 18]
    # }

    return inputs


class SRLDataFlow(DataFlow):
  """DataFlow implementation based on SRL task."""
  def __init__(self, config, task_name, tokenizers, label_mapping):
    super(SRLDataFlow, self).__init__(config, task_name, tokenizers, label_mapping)
    self._inv_label_mapping = {v: k for k, v in label_mapping.items()}
    self.pos_tag_label_mapping = json.load(open("data/preprocessed_data/pos_tag.json"))
    self.predicate_sense_label_mapping = json.load(open("data/preprocessed_data/predicate_sense_label_mapping.json"))
    self.use_gold_predicate = hasattr(config, "srl_use_gold_predicate") and config.srl_use_gold_predicate
    self.use_gold_argument = hasattr(config, "srl_use_gold_argument") and config.srl_use_gold_argument
    if hasattr(config, "predicate_reveal_method"):
      self.predicate_reveal_method = config.predicate_reveal_method
    else:
      self.predicate_reveal_method = None

  @property
  def example_class(self):
    """Return SRLExample class"""

    return SRLExample

  @property
  def minibatch_class(self):
    """Return SRLMiniBatch class"""

    return SRLMiniBatch

  def process_example(self, example: SRLExample):
    """Process SRL example with extra arguments.
    Including label_mapping and srl_label_format

    """
    example.process(tokenizers=self.tokenizers,
                    label_mapping=self.label_mapping,
                    pos_tag_label_mapping=self.pos_tag_label_mapping,
                    predicate_sense_mapping=self.predicate_sense_label_mapping,
                    label_format=self.config.srl_label_format,
                    use_gold_predicate=self.use_gold_predicate,
                    use_gold_argument=self.use_gold_argument,
                    predicate_reveal_method=self.predicate_reveal_method)

  def convert_examples_to_features(self, examples: List[SRLExample]):
    """
    TODO: The implementation is based on BERT model.
    The convert for other models will be implemeted.

    """
    examples: List[SRLExample]
    label_format = self.config.srl_label_format
    features = []

    # BERT based variables
    max_token_length = max([example.len for example in examples])
    if label_format is not None:
      max_label_length = max([example.label_len for example in examples])
    else:
      max_label_length = None

    max_arg_candidate_length = max(
        [len(example.enumerated_span_candidates) for example in examples])
    max_predicate_candidate_length = max(
        [len(example.predicate_candidates) for example in examples])

    # Label embedding variables
    try:
      max_label_candidate_length = max([
          max([len(label_candidates) for label_candidates in example.label_candidates])
        for example in examples])
    except:
      max_label_candidate_length = None

    # Classical based variable
    # _max_token_length = max([example._len for example in examples])
    # if label_format is not None:
    #   _max_label_length = max([example._label_len for example in examples])
    # else:
    #   _max_label_length = None
    #
    # _max_arg_candidate_length = max([
    #   len(example._enumerated_span_candidates) for example in examples])
    # _max_predicate_candidate_length = max(
    #   [len(example._predicate_candidates) for example in examples])

    for idx, example in enumerate(examples):
      # Bert based feature process
      padding = [0] * (max_token_length - example.len)
      input_ids = example.input_ids + padding
      input_mask = example.input_mask + padding
      segment_ids = example.segment_ids + padding
      is_head = example.is_head + [2] * (max_token_length - len(example.is_head))
      # Because we do not have is_head for extra predicate part, so we need to recalculate the
      # padding for is_head

      arg_candidates = example.enumerated_span_candidates + [
          (1, 0)
      ] * (max_arg_candidate_length - len(example.enumerated_span_candidates))
      predicate_candidates = example.predicate_candidates + [
          (1, 0)
      ] * (max_predicate_candidate_length - len(example.predicate_candidates))

      if example.predicate_index is not None:
        predicate_idx = example.predicate_idx
      else:
        predicate_idx = None

      # For label processing, there are three choice
      # 1. None if it is for prediction
      # 2. Sequence Labeling format, then it is just List[str]. This is predicate based.
      # 3. Span format, then it is List[Tuple[int, int, int, int, int]]
      if label_format is None or self.config.mode in ["deployment"]:
        label_ids = None
        arg_candidate_label_ids = None
        predicate_candidate_label_ids = None
      elif label_format == SRL_LABEL_SPAN_BASED:
        # span (1, 0) is a invalid span. We use that as padding
        label_ids = example.label_ids + [
            (1, 0, 1, 0, example.label_padding_id)
        ] * (max_label_length - example.label_len)

        arg_candidate_label_ids = example.arg_span_label_ids + [
          0] * (max_arg_candidate_length - len(example.enumerated_span_candidates))
        predicate_candidate_label_ids = example.pred_span_label_ids + [
          0] * (max_predicate_candidate_length - len(example.predicate_candidates))


      elif label_format == SRL_LABEL_SEQ_BASED:
        label_ids = example.label_ids + [example.label_padding_id] * (
            max_label_length - example.label_len)
        arg_candidate_label_ids = None
        predicate_candidate_label_ids = None
      else:
        raise ValueError(
            "The label format {} is not supported. Choice: {} | {}".format(
                label_format, SRL_LABEL_SEQ_BASED, SRL_LABEL_SPAN_BASED))

      # POS tag joint learning

      if example.pos_tag_label is not None:
        pos_tag_label_ids = example.pos_tag_label_ids + [example.pos_tag_label_padding_id] * (max_token_length - example.len)
      else:
        pos_tag_label_ids = None

      # Label embeddings
      # It will be a matrix. The first dim will be the length of the sentence (max), the second dim will
      # be the length of the label candidates (max)
      if max_label_candidate_length is not None:
        label_candidates_list = []
        label_candidates_mask_list = []
        
        for label_candidates in example.label_candidates:
          padded_label_candidates = label_candidates + [0] * (max_label_candidate_length - len(label_candidates))
          padded_label_candidates_mask = [1] * len(label_candidates) + [0] * (max_label_candidate_length - len(label_candidates))
          label_candidates_list.append(padded_label_candidates)
          label_candidates_mask_list.append(padded_label_candidates_mask)
        while len(label_candidates_list) < max_predicate_candidate_length:
          # We assue the predicate length is same as the original sentence length without BPE tokenization
          label_candidates_list.append([0] * max_label_candidate_length)
          label_candidates_mask_list.append([0] * max_label_candidate_length)
      else:
        label_candidates_list = None
        label_candidates_mask_list = None


      # Classical based feature process
      # _input_token_ids = example._input_token_ids + [0] * (_max_token_length - example._len)
      # _arg_candidates = example._enumerated_span_candidates + [
      #   (1, 0)
      # ] * (_max_arg_candidate_length - len(example._enumerated_span_candidates))
      # _predicate_candidates = example._predicate_candidates + [
      #   (1, 0)
      # ]* (_max_predicate_candidate_length - len(example._predicate_candidates))
      #
      # if label_format is None:
      #   _label_ids = None
      #
      # elif label_format == SRL_LABEL_SPAN_BASED:
      #   _label_ids = example._label_ids + [
      #     (1, 0, 1, 0, example.label_padding_id)
      #   ] * (_max_label_length - example._label_len)
      #
      # elif label_format == SRL_LABEL_SEQ_BASED:
      #   _label_ids = example._label_ids + [example.label_padding_id] * (
      #     _max_label_length - example._label_len)
      # else:
      #   raise ValueError(
      #     "The label format {} is not supported. Choice: {} | {}".format(
      #       label_format, SRL_LABEL_SEQ_BASED, SRL_LABEL_SPAN_BASED))

      features.append(
          SRLFeature(input_ids=input_ids,
                     input_mask=input_mask,
                     segment_ids=segment_ids,
                     is_head=is_head,
                     label_ids=label_ids,
                     arg_candidates=arg_candidates,
                     predicate_candidates=predicate_candidates,
                     arg_candidate_label_ids = arg_candidate_label_ids,
                     predicate_candidate_label_ids = predicate_candidate_label_ids,
                     label_candidates = label_candidates_list,
                     label_candidates_mask = label_candidates_mask_list,
                     pos_tag_label_ids=pos_tag_label_ids,
                     predicate_idx=predicate_idx,
                     _input_token_ids=None, # _input_token_ids,
                     _label_ids=None, #_label_ids,
                     _arg_candidates=None, # _arg_candidates,
                     _predicate_candidates=None, #_predicate_candidates,
                     _token_length=0# example._len
                     ))
    return features

  def decode_to_labels(self, preds, mb: MiniBatch):
    labels = []
    for example, pred_logits in zip(mb.examples, preds[mb.task_name]["logits"]):
      if example.is_valid:
        preds_tags = pred_logits.argmax(-1).data.cpu().numpy()
        span_preds, pred_labels = get_span_labels(
          sentence_tags=preds_tags,
          is_head=example.is_head,
          segment_id=example.segment_ids,
          inv_label_mapping=self._inv_label_mapping)
        span_preds = set(filter(lambda item: item[0] != example.predicate_index, span_preds))
        # Because the get_span_labels is inclusive
        span_preds = [(x[0], x[1]+1, x[2]) for x in span_preds]
      else:
        span_preds = []

      labels.append(span_preds)
    return labels


from relogic.logickit.scorer.scorer import Scorer
import abc


class SpanLevelScorer(Scorer, metaclass=abc.ABCMeta):
  def __init__(self):
    super(SpanLevelScorer, self).__init__()
    self._examples = []
    self._boundaries = []
    self._preds = []

  def update(self, mbs, predictions, loss, extra):
    super(SpanLevelScorer, self).update(mbs, predictions, loss, extra)
    span_boundary_prediction, span_label_prediction = predictions
    start_index, end_index = span_boundary_prediction
    if len(mbs.examples) != len(start_index):
      print("The number of examples is not same as pred examples {} vs {} vs {} vs {}".format(len(mbs.examples),
        len(start_index), len(end_index), len(span_label_prediction)))
      for example in mbs.examples:
        print(example.__dict__)
      print(span_boundary_prediction, span_label_prediction)
      exit()
    for example, boundary_preds_start, boundary_preds_end, label_preds in zip(mbs.examples, start_index, end_index, span_label_prediction):
      self._examples.append(example)
      self._boundaries.append(zip(boundary_preds_start.data.cpu().numpy(), boundary_preds_end.data.cpu().numpy()))
      self._preds.append(label_preds)


from relogic.logickit.tasks.task import Task
from relogic.logickit.dataset.labeled_data_loader import LabeledDataLoader
from relogic.logickit.modules.parallel_mapping_module import ParallelMappingModule
from relogic.logickit.modules.select_index_module import SelectIndexModule
from relogic.logickit.modules.mixsent_alignment import MixSentAlignmentModule
from relogic.logickit.modules.gen_repr import GenRepr
from relogic.logickit.base.constants import (PARALLEL_MAPPING_TASK,
  PARALLEL_TEACHER_STUDENT_TASK, MIXSENT_TASK, LANGUAGE_IDENTIFICATION_IR, LANGUAGE_IDENTIFICATION_SEQ)
from relogic.logickit.scorer.distance_scorer import DistanceScorer



class Unsupervised(Task):
  def __init__(self, config, name, tokenizer=None):
    super(Unsupervised, self).__init__(
      config, name, LabeledDataLoader(config, name, tokenizer))
    self.config = config

  def get_module(self):
    if self.name in [PARALLEL_MAPPING_TASK]:
      return ParallelMappingModule(self.config, self.name)
    if self.name in [PARALLEL_TEACHER_STUDENT_TASK]:
      return SelectIndexModule(self.config, self.name)
    if self.name in [MIXSENT_TASK]:
      return MixSentAlignmentModule(self.config, self.name)
    if self.name in [LANGUAGE_IDENTIFICATION_IR, LANGUAGE_IDENTIFICATION_SEQ]:
      return GenRepr(self.config, self.name)
    else:
      raise ValueError("Can not find task name {}".format(self.name))

  def get_scorer(self, dump_to_file=None):
    if self.name in [LANGUAGE_IDENTIFICATION_IR]:
      return None
    return DistanceScorer()


from relogic.logickit.tasks.task import Task
from relogic.logickit.dataset.labeled_data_loader import LabeledDataLoader
from relogic.logickit.modules.matching_module import MatchingModule
from relogic.logickit.modules.classification_module import ClassificationModule
from relogic.logickit.modules.representation_module import RepresentationModule
from relogic.logickit.modules.rel_extraction_module import RelExtractionModule
from relogic.logickit.modules.agg_matching_module import AggMatchingModule
from relogic.logickit.modules.ir_matching import IRMatchingModule
from relogic.logickit.scorer.ranking_scorer import RecallScorer, CartesianMatchingRecallScorer, RetrievalScorer
from relogic.logickit.scorer.classification_scorers import RelationF1Scorer, MultiClassAccuracyScorer
from relogic.logickit.modules.sam_cnn import SamCNN
from relogic.logickit.base.constants import IR_TASK, PAIRWISE_TASK, SINGLETON, ENTITY_TYPE_CLASSIFICATION, DOCIR_TASK, IR_SAMCNN_TASK


class Classification(Task):
  def __init__(self, config, name, tokenizer=None):
    super(Classification, self).__init__(
      config, name, LabeledDataLoader(config, name, tokenizer))
    self.n_classes = len(set(self.loader.label_mapping.values()))

  def get_module(self):
    if self.name in ["matching"]:
      return MatchingModule(self.config, self.name, self.n_classes)
    elif self.name in [IR_SAMCNN_TASK]:
      return SamCNN(self.config, self.name, self.n_classes)
    elif self.name.startswith(IR_TASK):
      if self.config.word_level_interaction:
        return IRMatchingModule(self.config, self.name, self.n_classes)
      else:
        return MatchingModule(self.config, self.name, self.n_classes)
    elif self.name.startswith(DOCIR_TASK):
      return AggMatchingModule(self.config, self.name, self.n_classes)
    elif self.name in ["pair_matching", PAIRWISE_TASK]:
      return RepresentationModule(self.config, self.name, self.config.repr_size)
    elif self.name in ["rel_extraction", SINGLETON]:
      return RelExtractionModule(self.config, self.name, self.n_classes)
    elif self.name in [ENTITY_TYPE_CLASSIFICATION]:
      return ClassificationModule(self.config, self.name, self.n_classes)


  def get_scorer(self, dump_to_file=None):
    if self.name in ["matching"]:
      return RecallScorer(self.loader.label_mapping, topk=self.config.topk, dump_to_file=dump_to_file)
    elif self.name.startswith(IR_TASK) or self.name.startswith(DOCIR_TASK):
      return RetrievalScorer(
        self.loader.label_mapping,
        qrels_file_path=self.config.qrels_file_path if isinstance(self.config.qrels_file_path, str)
            else self.config.tasks[self.name]["qrels_file_path"],
        dump_to_file=dump_to_file,
        regression=self.config.regression)
    elif self.name in ["rel_extraction"]:
      return RelationF1Scorer(self.loader.label_mapping, dump_to_file=dump_to_file)
    elif self.name in ["pair_matching", PAIRWISE_TASK]:
      return CartesianMatchingRecallScorer(
        topk=self.config.topk,
        qrels_file_path=self.config.qrels_file_path if isinstance(self.config.qrels_file_path, str)
          else self.config.tasks[self.name]["qrels_file_path"],
        dump_to_file=dump_to_file)
    elif self.name in [ENTITY_TYPE_CLASSIFICATION]:
      return MultiClassAccuracyScorer(self.loader.label_mapping, dump_to_file=dump_to_file, dataflow=self.loader.dataflow)



from relogic.logickit.scorer.scorer import Scorer
from itertools import groupby
import collections
import json
from relogic.logickit.scripts import squad_eval
import os

RawResult = collections.namedtuple("RawResult", ["start_logits", "end_logits"])

class SpanExtractionScorer(Scorer):
  def __init__(self, dataset, gold_answer_file, null_score_diff_threshold, dump_to_file=None):
    super(SpanExtractionScorer, self).__init__()
    self.guid_to_example = {}
    self.dataset_name = dataset
    # dataset_json = json.load(open("data/preprocessed_data/{}_dev.json".format(dataset)))
    dataset_json = json.load(open(gold_answer_file))
    self.dataset = dataset_json['data']
    self.null_score_diff_threshold = null_score_diff_threshold

    if dump_to_file:
      self.dump_to_file_path = os.path.join(dump_to_file["output_dir"], dump_to_file["task_name"] + "_dump.json")
      self.dump_to_null_odds_path = os.path.join(dump_to_file["output_dir"], dump_to_file["task_name"] + "_null_odds.json")
      self.dump_to_span_scores_path = os.path.join(dump_to_file["output_dir"], dump_to_file["task_name"] + "_span_scores.json")


  def update(self, mb, predictions, loss, extra_args):
    # Basically one example can map several features
    # How to map it back ?
    # 1. Build guid to example mapping
    # 2. Map the input_feature to example using guid
    for example in mb.examples:
      self.guid_to_example[example.guid] = example
    pair_group = groupby(zip(mb.input_features, predictions[0], predictions[1]), key=lambda pair: pair[0].guid)
    pair_group = {k: list(v) for k, v in pair_group}
    # {guid: list of pairs}
    for guid in pair_group.keys():
      raw_features = []
      raw_results = []
      for item in pair_group[guid]:
        raw_features.append(item[0])
        raw_results.append(RawResult(start_logits=item[1], end_logits=item[2]))
      self.guid_to_example[guid].write_predictions(
        raw_features=raw_features, 
        raw_results=raw_results,
        n_best_size=20,
        with_negative=mb.task_name in ["squad20"])

  def _get_results(self):
    assert self.dump_to_file_path is not None

    if self.dump_to_file_path:
      self.dump_to_file_handler = open(self.dump_to_file_path, 'w')
      self.dump_to_null_odds_handler = open(self.dump_to_null_odds_path, 'w')
      self.dump_to_span_scores_handler = open(self.dump_to_span_scores_path, 'w')

    prediction_json = {}

    for example in self.guid_to_example.values():
      prediction_json[example.guid] = example.prediction
    json.dump(prediction_json, self.dump_to_file_path)
    if self.dataset_name in ["squad20"]:
      scores_diff_json = {}
      for example in self.guid_to_example.values():
        scores_diff_json[example.guid] = example.scores_diff
      json.dump(scores_diff_json, self.dump_to_null_odds_handler)

      span_scores_json = {}
      for example in self.guid_to_example.values():
        span_scores_json[example.guid] = example.span_score
      json.dump(span_scores_json, self.dump_to_span_scores_handler)

    return squad_eval(self.dataset_name)(self.dataset, prediction_json,
              self.null_score_diff_threshold)

  def get_loss(self):
    return 0


import tensorflow as tf
from relogic.tpukit.training.trainer import Trainer
flags = tf.flags

FLAGS = flags.FLAGS

## Required parameters
flags.DEFINE_string(
    "data_dir", None,
    "The input data dir. Should contain the .tsv files (or other data files) "
    "for the task.")

flags.DEFINE_string(
    "bert_config_file", None,
    "The config json file corresponding to the pre-trained BERT model. "
    "This specifies the model architecture.")

flags.DEFINE_string("task_name", None, "The name of the task to train.")

flags.DEFINE_string("vocab_file", None,
                    "The vocabulary file that the BERT model was trained on.")

flags.DEFINE_string(
    "output_dir", None,
    "The output directory where the model checkpoints will be written.")

## Other parameters

flags.DEFINE_string(
    "init_checkpoint", None,
    "Initial checkpoint (usually from a pre-trained BERT model).")

flags.DEFINE_bool(
    "do_lower_case", True,
    "Whether to lower case the input text. Should be True for uncased "
    "models and False for cased models.")

flags.DEFINE_integer(
    "max_seq_length", 128,
    "The maximum total input sequence length after WordPiece tokenization. "
    "Sequences longer than this will be truncated, and sequences shorter "
    "than this will be padded.")

flags.DEFINE_bool("do_pretraining", False, "Whether to run pretraining")

flags.DEFINE_bool("do_train", False, "Whether to run training.")

flags.DEFINE_bool("do_eval", False, "Whether to run eval on the dev set.")

flags.DEFINE_bool(
    "do_predict", False,
    "Whether to run the model in inference mode on the test set.")

flags.DEFINE_integer("train_batch_size", 32, "Total batch size for training.")

flags.DEFINE_integer("eval_batch_size", 8, "Total batch size for eval.")

flags.DEFINE_integer("predict_batch_size", 8, "Total batch size for predict.")

flags.DEFINE_integer("eval_dev_every", 50, "Do validation for every steps")

flags.DEFINE_float("learning_rate", 5e-5, "The initial learning rate for Adam.")

flags.DEFINE_float("num_train_epochs", 3.0,
                   "Total number of training epochs to perform.")

flags.DEFINE_integer("num_train_steps", 100000, "Number of training steps.")

flags.DEFINE_integer("num_warmup_steps", 10000, "Number of warmup steps.")

flags.DEFINE_float(
    "warmup_proportion", 0.1,
    "Proportion of training to perform linear learning rate warmup for. "
    "E.g., 0.1 = 10% of training.")

flags.DEFINE_integer("save_checkpoints_steps", 1000,
                     "How often to save the model checkpoint.")

flags.DEFINE_integer("iterations_per_loop", 1000,
                     "How many steps to make in each estimator call.")

flags.DEFINE_bool("use_tpu", False, "Whether to use TPU or GPU/CPU.")

tf.flags.DEFINE_string(
    "tpu_name", None,
    "The Cloud TPU to use for training. This should be either the name "
    "used when creating the Cloud TPU, or a grpc://ip.address.of.tpu:8470 "
    "url.")

tf.flags.DEFINE_string(
    "tpu_zone", None,
    "[Optional] GCE zone where the Cloud TPU is located in. If not "
    "specified, we will attempt to automatically detect the GCE project from "
    "metadata.")

tf.flags.DEFINE_string(
    "gcp_project", None,
    "[Optional] Project name for the Cloud TPU-enabled project. If not "
    "specified, we will attempt to automatically detect the GCE project from "
    "metadata.")

tf.flags.DEFINE_string("master", None, "[Optional] TensorFlow master URL.")

flags.DEFINE_integer(
    "num_tpu_cores", 8,
    "Only used if `use_tpu` is True. Total number of TPU cores to use.")

tf.logging.set_verbosity(tf.logging.INFO)

trainer = Trainer(FLAGS)
trainer.train()


import json
from types import SimpleNamespace

import relogic.utils.crash_on_ipy
from relogic.logickit.dataflow import SRLDataFlow
from relogic.logickit.tokenizer.fasttext_tokenization import FasttextTokenizer
from transformers.tokenization_bert import BertTokenizer
from relogic.structures.sentence import Sentence

config = SimpleNamespace(
  **{
    "buckets": [(0, 15), (15, 40), (40, 450)],
    "max_seq_length": 450,
    "label_mapping_path": "data/preprocessed_data/srl_conll05_label_mapping.json",
    "srl_label_format": "srl_label_span_based"
  })

tokenizers = {
  "BPE": BertTokenizer.from_pretrained("bert-base-cased"),
  "Fasttext": FasttextTokenizer.from_pretrained("wiki-news-300d-1M")
}

dataflow = SRLDataFlow(task_name="joint_srl",
                       config=config,
                       tokenizers=tokenizers,
                       label_mapping=json.load(open(
                         config.label_mapping_path)))

structures = [Sentence(text="I went to Paris yesterday .")]
dataflow.update_with_structures(structures)

examples = [{
  "tokens": ["I", "went", "to", "Paris", "yesterday", "."],
  "labels": [(1, 2, "went", 0, 1, "I", "A0"),
             (1, 2, "went", 3, 4, "Paris", "A1")],
  "label_candidates": [[0], [0, 1, 2, 3], [0], [0], [0], [0]],
  "pos_tag": ["NN", "VB", "NN", "NN", "NN", "NN"]
}, {
  "tokens": ["But", "I", "didn't", "find", "you", "lol", "."],
  "labels": [(3, 4, "find", 0, 1, "But", "AM-DIS")],
  "label_candidates": [[0], [0], [0], [5, 6, 7, 8, 9], [0], [0], [0]],
  "pos_tag": ["NN", "VB", "VB", "VB", "VB", "NN"]
}]
dataflow.update_with_jsons(examples)

config = SimpleNamespace(
  **{
    "buckets": [(0, 15), (15, 40), (40, 450)],
    "max_seq_length": 450,
    "label_mapping_path": "data/preprocessed_data/srl_conll05_BIO_label_mapping.json",
    "srl_label_format": "srl_label_seq_based",
    "predicate_reveal_method": "srl_predicate_extra_surface"
  })

dataflow = SRLDataFlow(task_name="srl",
                       config=config,
                       tokenizers=tokenizers,
                       label_mapping=json.load(open(
                         config.label_mapping_path)))
examples = [{
  "tokens": ["I", "went", "to", "Paris", "yesterday", "."],
  "labels": [(1, 2, "went", 0, 1, "I", "A0"),
             (1, 2, "went", 3, 4, "Paris", "A1")],
  "label_candidates": [[0], [0, 1, 2, 3], [0], [0], [0], [0]],
  "pos_tag": ["NN", "VB", "NN", "NN", "NN", "NN"],
  "predefined_predicate": "went"
}, {
  "tokens": ["But", "I", "didn't", "find", "you", "lol", "."],
  "labels": [(3, 4, "find", 0, 1, "But", "AM-DIS")],
  "label_candidates": [[0], [0], [0], [5, 6, 7, 8, 9], [0], [0], [0]],
  "pos_tag": ["NN", "VB", "VB", "VB", "VB", "NN"],
  "predefined_predicate": "find"
}]
dataflow.update_with_jsons(examples)


for mb in dataflow.get_minibatches(minibatch_size=2):
  print(mb)

raise NotImplementedError("You can start to play with data")



from relogic.logickit.utils import get_span_labels

import argparse
from itertools import groupby
import json

from tqdm import tqdm
import spacy
from spacy.tokens import Doc

class WhitespaceTokenizer(object):
    def __init__(self, vocab):
        self.vocab = vocab

    def __call__(self, text):
        words = text.split(' ')
        # All tokens 'own' a subsequent space character in this tokenizer
        spaces = [True] * len(words)
        return Doc(self.vocab, words=words, spaces=spaces)

nlp = spacy.load("en_core_web_sm", disable=["parser", "ner"])
nlp.tokenizer = WhitespaceTokenizer(nlp.vocab)

def main(json_file, dump_file):
  examples = []
  with open(json_file) as fin:
    for line in fin:
      example = json.loads(line)
      example["key"] = " ".join(example["tokens"])
      examples.append(example)

  agg_examples = segmentation(examples)
  with open(dump_file, 'w') as fout:
    for example in agg_examples:
      fout.write(json.dumps(example) + "\n")


def segmentation(examples):
  groups = groupby(examples, key=lambda x: x["key"])
  agg_examples = []
  for group in tqdm(groups):
    text = group[0].split()
    segment_label = ["I"] * len(text)
    segment_label[0] = "B"
    sent = nlp(" ".join(text))
    for idx, t in enumerate(sent):
      if t.pos_ == "ADP" or t.pos_ == "PUNCT":
        segment_label[idx] = 'B'
        if idx + 1 < len(text):
          segment_label[idx + 1] = 'B'
    for sub_item in group[1]:
      spans, _ = get_span_labels(sub_item["labels"], ignore_label=[])
      for span in spans:
        segment_label[span[0]] = 'B'
        if len(text) > span[1] + 1:
          segment_label[span[1] + 1] = 'B'

    example = {}
    example["tokens"] = text
    example["segment_labels"] = segment_label
    agg_examples.append(example)
  return agg_examples


if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("--json_file")
  parser.add_argument("--dump_file")
  args = parser.parse_args()
  main(args.json_file, args.dump_file)


import argparse
import os
import json
from collections import defaultdict


def convert_into_group(input_folder, output_folder):
  if not os.path.exists(output_folder):
    os.makedirs(output_folder)
  for split in ["train", "dev", "test"]:

    with open(os.path.join(input_folder, "{}.json".format(split))) as fin:
      print("Processing {}".format(fin.name))
      data = json.load(fin)
      data_collection = defaultdict(list)
      for example in data:
        text = " ".join(example["token"])
        data_collection[text].append(example)
      print("There are {} examples, and {} sentences".format(
        len(data), len(data_collection)))
    with open(os.path.join(output_folder, "{}.json".format(split)), "w") as fout:
      print("Dumping results to {}".format(fout.name))
      for sentence, group in data_collection.items():
        tokens = sentence.split()
        new_example = {
          "sentence": sentence,
          "masked_sentence": None,
          "subj_index": [],
          "obj_index": [],
          "subj_type": [],
          "obj_type": [],
          "subj_text": [],
          "obj_text": [],
          "relation": []}
        entity_subj_spans = []
        entity_obj_spans = []
        relations = []
        for example in group:
          start_index = int(example["subj_start"])
          end_index = int(example["subj_end"]) + 1
          span = (start_index, end_index, example["subj_type"], " ".join(tokens[start_index: end_index]))
          entity_subj_spans.append(span)
          start_index = int(example["obj_start"])
          end_index = int(example["obj_end"]) + 1
          span = (start_index, end_index, example["obj_type"], " ".join(tokens[start_index: end_index]))
          entity_obj_spans.append(span)
          relations.append(example["relation"])
        entity_spans = sorted(list(set(entity_subj_spans + entity_obj_spans)), key=lambda x: x[0])
        new_sentence = []
        start_index = 0
        span_to_index = {}
        for span in entity_spans:
          new_sentence.extend(tokens[start_index: span[0]])
          span_to_index[span] = len(new_sentence)
          new_sentence.append("[{}]".format(span[2]))
          start_index = span[1]
        new_sentence.extend(tokens[start_index:])
        for span in entity_subj_spans:
          new_example["subj_index"].append(span_to_index[span])
          new_example["subj_type"].append(span[2])
          new_example["subj_text"].append(span[3])
        for span in entity_obj_spans:
          new_example["obj_index"].append(span_to_index[span])
          new_example["obj_type"].append(span[2])
          new_example["obj_text"].append(span[3])
        for relation in relations:
          new_example["relation"].append(relation)
        new_example["masked_sentence"] = new_sentence
        fout.write(json.dumps(new_example) + "\n")

def convert(input_folder, output_folder):
  if not os.path.exists(output_folder):
    os.makedirs(output_folder)

  for split in ["train", "dev", "test"]:
    examples = []
    with open(os.path.join(input_folder, "{}.json".format(split))) as fin:
      print("Processing {}".format(fin.name))
      data = json.load(fin)
      for example in data:
        processed = {}
        tokens = example["token"]
        processed["text"] = " ".join(tokens)

        start_index = int(example["subj_start"])
        end_index = int(example["subj_end"]) + 1
        processed["subj_text"] = " ".join(tokens[start_index: end_index])
        processed["subj_span"] = (start_index, end_index)

        start_index = int(example["obj_start"])
        end_index = int(example["obj_end"]) + 1
        processed["obj_text"] = " ".join(tokens[start_index: end_index])
        processed["obj_span"] = (start_index, end_index)

        processed["subj_type"] = example["subj_type"]
        processed["obj_type"] = example["obj_type"]

        processed["label"] = example["relation"]
        examples.append(processed)
    with open(os.path.join(output_folder, "{}.json".format(split)), "w") as fout:
      for example in examples:
        fout.write(json.dumps(example) + "\n")

if __name__ == "__main__":
  parser = argparse.ArgumentParser()
  parser.add_argument("--input_folder", type=str)
  parser.add_argument("--output_folder", type=str)
  args = parser.parse_args()
  # convert_into_group(args.input_folder, args.output_folder)
  convert(args.input_folder, args.output_folder)


from typing import Dict, List, Union

from relogic.structures.sentence import Sentence
from relogic.structures.span import Span
from relogic.structures.structure import Structure
from relogic.structures.document import Document
from relogic.structures.linkage_candidate import LinkageCandidate

from relogic.graphkit.utils.similarity_function import jaccard_similarity
from relogic.utils.file_utils import cached_path, RELOGIC_CACHE
import os
from zipfile import ZipFile
import logging

logger = logging.getLogger(__name__)  # pylint: disable=invalid-name

"""
A quick hack
"""
from jnius import autoclass
JString = autoclass('java.lang.String')
JSearcher = autoclass('io.anserini.search.KGSearcher')

"""
End of quick hack
"""

INDEX_PATHS = {
  "en_wikipedia": "https://git.uwaterloo.ca/p8shi/data-server/raw/master/entity_linking_index_en.zip",
  "zh_baike": "https://git.uwaterloo.ca/p8shi/data-server/raw/master/entity_linking_index_zh.zip"
}

INDEX_LANGUAGE = {
  "en_wikipedia": "en",
  "zh_baike": "zh"
}

class SimpleEntityLinker(object):
  """Two step linking.
  
  Args:
    paths (Dict): The key is the name for the retriever, and value is the directory path to
      the Lucene index.
  """
  def __init__(self, index_names: List, index_paths: Dict = None, index_language: Dict = None):

    self.index_names = index_names
    self.retrievers = {}
    if index_paths is not None and index_language is not None:
      INDEX_PATHS.update(index_paths)
      INDEX_LANGUAGE.update(index_language)
    for index_name in index_names:
      index_zip_or_dir_path = cached_path(INDEX_PATHS[index_name], cache_dir=RELOGIC_CACHE)
      if os.path.isdir(index_zip_or_dir_path):
        index_path = index_zip_or_dir_path
      else:
        index_path = index_zip_or_dir_path + "." + index_name
      if not os.path.exists(index_path):
        with ZipFile(index_zip_or_dir_path, 'r') as zipObj:
          zipObj.extractall(index_path)
          logger.info("Extract Index {} to {}".format(INDEX_PATHS[index_name], index_path))
      self.retrievers[index_name] = JSearcher(JString(index_path))
      self.retrievers[index_name].setLanguage(INDEX_LANGUAGE[index_name])

  def entity_retrieval(self, mention: Span, name: str, candidate_size: int = 20):
    hits = self.retrievers[name].search(
      JString(mention.text.encode("utf-8")), JString("contents"), True, None, candidate_size)
    # Retrieve entity linking results from Anserini.
    # The data structure follows the definition in Anserini.
    # LinkageCandidate.from_hit need to follow the changes in Anserini.
    for i in range(len(hits)):
      mention.add_linkage_candidate(LinkageCandidate.from_hit(hits[i]))

  def rank(self, mention: Span, epsilon: float = 0.5):
    """Operate based on span, and change the results in position.
    1. Sorted by retrieval score
    2. Given score range [max_score - epsilon, max_score], aggregate the URIs
      from prior or/and alias_of
    3. Convert the URI to string (process the URI or use the label of URI), and 
      rank with the string # A quick hack for DBpedia or Wikidata.

    Args:
      mention (:data:`Span`): Mention with linkage candidates for ranking
      epsilon (float): Control the first layer score range from
        [max_score - epsilon, max_score]
    """
    if mention.linkage_candidates:
      mention.linkage_candidates = sorted(mention.linkage_candidates,
        key=lambda x: self.similarity(x.text, mention.text), reverse=True)
    else:
      return
    
    max_score = mention.linkage_candidates[0].score

    # Aggregate first layer
    # Idempotence operation
    mention.first_layer_prior = {}
    for candidate in mention.linkage_candidates:
      if abs(candidate.score - max_score) < epsilon:
        # Go over the priors
        for uri, count in candidate.prior.items():
          if uri:
            mention.first_layer_prior[uri] = mention.first_layer_prior.get(uri, 0) + count
        # Also consider the alias_of
        for uri in candidate.alias_of:
          if uri:
            mention.first_layer_prior[uri] = mention.first_layer_prior.get(uri, 0) + 1
          # TODO: Here 1 is a hyper parameter

    # Sorted by the prior
    sorted_uris = sorted(mention.first_layer_prior.items(), key= lambda x: x[1], reverse=True)
    # TODO: Consider the two ranking together with weighted sum

    # Hard selection for the exact match
    # TODO: This is hard coded for DBpedia or Wikidata
    mention_text_to_uri = mention.text.lower().replace(" ", "_")
    fixed_sequence = []
    for idx, item in enumerate(sorted_uris):
      # item = (uri, count)
      if mention_text_to_uri == item[0].lower():
        fixed_sequence = [sorted_uris[idx]] + sorted_uris[:idx] + sorted_uris[idx+1:]
        break
    if len(fixed_sequence) > 0:
      mention.ranked_uris = fixed_sequence
    else:
      mention.ranked_uris = sorted_uris

  @staticmethod
  def similarity(x, y):
    return jaccard_similarity(x.lower().split(), y.lower().split())

  def global_inference(self, inputs: Structure):
    """The global inference is to deal with mention boundary overlap.
    Due to imperfect entity detection, there will be some uncertain entities, causing
      the overlap of some entities. For example, in sentence 'Who painted The Storm on 
      the Sea of Galilee', 'The Storm on the Sea of Galilee' and 'The Storm' are
      extracted as mentions at the same time due to different entity mention sources are
      aggregated. With global inference, duplicate/incorrect entity such as 'The Storm' 
      will be removed because it is full substring of 'The Storm on the Sea of Galilee'.
      This process can not be done before linking, because we do not know which mention is 
      better solely based on the surface form of the mention. With preliminary entity
      linking results, we can resolve these better.
    """
    pass


  def link_span(self, span: Span, name: str):
    self.entity_retrieval(span, name=name)
    self.rank(span)

  def link(self, inputs: Union[str, Structure, List[Structure]], name:str):
    """Linking method.
    If the inputs is Sentence, then two steps linking is operated.
    If the inputs is a Span, then candidate retrieval is operated 
      without global inference.
    """
    if isinstance(inputs, str):
      span = Span(text=inputs)
      self.link_span(span, name=name)
      return span
    elif isinstance(inputs, Structure):
      if isinstance(inputs, Document):
        self.link(inputs.sentences, name=name)
      elif isinstance(inputs, Sentence):
        for span in inputs.spans:
          self.link_span(span, name=name)
        self.global_inference(inputs)
      elif isinstance(inputs, Span):
        self.link_span(inputs, name=name)
      else:
        raise ValueError("Item type {} is not supported".format(type(inputs)))
    elif isinstance(inputs, List):
      for item in inputs:
        self.link(item, name=name)
    else:
      raise ValueError("Item type {} is not supported".format(type(inputs)))




import torch
import torch.nn as nn
from relogic.logickit.utils import utils
import torch.nn.functional as F

class IRMatchingModule(nn.Module):
  def __init__(self, config, task_name, n_classes):
    super().__init__()
    self.config = config
    self.task_name = task_name
    if self.config.regression:
      self.n_classes = 1
    else:
      self.n_classes = n_classes
    self.to_logits = nn.Sequential(
      nn.Linear(config.hidden_size * 2, config.hidden_size),
      nn.ReLU(),
      nn.Linear(config.hidden_size, self.n_classes)
    )

  def forward(self, *inputs, **kwargs):
    features = kwargs.pop("features")

    text_a_indices = kwargs.pop("text_a_indices")
    text_b_indices = kwargs.pop("text_b_indices")
    text_a_mask = (text_a_indices > 0).float()
    text_b_mask = (text_b_indices > 0).float()
    text_a_lengths = torch.sum(text_a_mask, dim=-1)
    text_b_lengths = torch.sum(text_b_mask, dim=-1)

    batch_size = features.size(0)
    query_token_size = text_a_indices.size(1)
    doc_token_size = text_b_indices.size(1)
    query_vector = utils.batched_index_select_tensor(features, text_a_indices)
    # select query vector based on the query indices
    doc_vector = utils.batched_index_select_tensor(features, text_b_indices)
    # select doc vector based on the doc indices

    exp_query_vector = query_vector.unsqueeze(2).repeat(1, 1, doc_token_size, 1)
    # expand the query vector into the shape (batch_size, max_query_token_size, max_doc_token_size, dim)
    exp_doc_vector = doc_vector.unsqueeze(1).repeat(1, query_token_size, 1, 1)
    # expand the doc vector into the shape (batch_size, max_query_token_size, max_doc_token_size, dim)
    sim = F.cosine_similarity(exp_query_vector.view(-1, query_vector.size(-1)),
      exp_doc_vector.view(-1, doc_vector.size(-1))).view(batch_size, query_token_size, doc_token_size)
    # convert both tensors into (-1, dim) shape and compute the consine similarity between these two tensors
    # and then convert it back into (batch_size, max_query_token_size, max_doc_token_size)

    scaled_doc_vector = sim.unsqueeze(-1) * exp_doc_vector
    # scale the document vector
    exp_text_b_mask = text_b_mask.unsqueeze(1).repeat(1, query_token_size, 1).unsqueeze(-1)
    exp_text_b_lengths = text_b_lengths.unsqueeze(-1).repeat(1, query_token_size).unsqueeze(-1)
    per_query_token_based_doc_repr = torch.sum(scaled_doc_vector * exp_text_b_mask, dim=-2) / (exp_text_b_lengths + 1)
    # compute the average repr of document, this repr is based each query token

    exp_text_a_mask = text_a_mask.unsqueeze(-1)
    exp_text_a_lengths = text_a_lengths.unsqueeze(-1)
    query_based_doc_repr = torch.sum(per_query_token_based_doc_repr * exp_text_a_mask, dim=-2) / (exp_text_a_lengths + 1)
    # aggregate document repr based on all query tokens

    query_vector_avg = torch.sum(query_vector * exp_text_a_mask , dim=-2) / (exp_text_a_lengths + 1)
    # aggregate the query repr

    feat = torch.cat([query_vector_avg, query_based_doc_repr], dim=-1)
    logits = self.to_logits(feat)
    return logits


READING_COMPREHENSION_TASKS = ["squad11", "squad20"]
AUTO_SPAN = ["auto_span_srl"]
NEVER_SPLIT = ("[UNK]", "[SEP]", "[PAD]", "[CLS]", "[MASK]",
              '[ORGANIZATION-SUBJ]', '[PERSON-OBJ]', '[PERSON-SUBJ]',
              '[ORGANIZATION-OBJ]', '[NUMBER-OBJ]', '[DATE-OBJ]',
              '[NATIONALITY-OBJ]', '[LOCATION-OBJ]', '[TITLE-OBJ]',
              '[CITY-OBJ]', '[MISC-OBJ]', '[COUNTRY-OBJ]',
              '[CRIMINAL_CHARGE-OBJ]', '[RELIGION-OBJ]',
              '[DURATION-OBJ]', '[URL-OBJ]', '[STATE_OR_PROVINCE-OBJ]',
              '[IDEOLOGY-OBJ]', '[CAUSE_OF_DEATH-OBJ]')

JOINT_SRL_TASK = "joint_srl"
PIPE_SRL_TASK = "pipe_srl"
SRL_LABEL_SPAN_BASED = "srl_label_span_based"
SRL_LABEL_SEQ_BASED = "srl_label_seq_based"
SPAN_REPR_KENTON_LEE = "kenton_lee"
SPAN_REPR_AVE_MAX = "ave_max"
SPAN_REPR_AVE = "ave"
SRL_PREDICATE_EXTRA_SURFACE = "srl_predicate_extra_surface"
PREDICATE_DETECTION_TASK = "predicate_detection"


ECP_TASK = "ecp"


# Language Code based on ISO 639-1 codes
LANG_EN = "en"
LANG_zH = "zh"

POINTWISE_TASK = "pointwise"
PAIRWISE_TASK = "pairwise"

IR_TASK = "ir"
DOCIR_TASK = "doc_ir"
IR_SIAMESE_TASK = "ir_siamese"
IR_SAMCNN_TASK = "ir_samcnn"


NER_TASK = "ner"
POS_TASK = "pos"
ENTITY_TYPE_CLASSIFICATION = "entity_type_classification"
SEQUENCE_LABELING_TASK = "sequence_labeling"
DEP_PARSING_TASK = "dep_parsing"
SEQUENCE_CLASSIFICATION_TASK = "sequence_classification"
PARALLEL_MAPPING_TASK = "parallel_mapping"
PARALLEL_TEACHER_STUDENT_TASK = "parallel_teacher_student"
MIXSENT_TASK = "mixsent"

LANGUAGE_IDENTIFICATION_IR = "language_identification_ir"
LANGUAGE_IDENTIFICATION_SEQ = "language_identification_seq"


SINGLETON = []
SIAMESE = [IR_SIAMESE_TASK, IR_SAMCNN_TASK]
TRIPLET = [PAIRWISE_TASK]

DISTILL_TASKS = ["parallel_teacher_student", "mixsent"]

ENCODING_TASK = "encoding"

AUXILIARY_TRAINING = "auxiliary_training"
ITERATIVE_TRAINING = "iterative_training"
ADVERSARIAL_TRAINING = "adversarial_training"

TRAIN_DISCRIMINATOR = "train_discriminator"
TRAIN_GENERATOR = "train_generator"

SKIP_LOSS_TASK = [LANGUAGE_IDENTIFICATION_IR, LANGUAGE_IDENTIFICATION_SEQ]
SKIP_EVAL_TASK = [LANGUAGE_IDENTIFICATION_IR, LANGUAGE_IDENTIFICATION_SEQ]





import torch.nn as nn

from relogic.logickit.base import utils
from relogic.logickit.inference.encoder import get_encoder
from relogic.logickit.inference.branching_encoder import BranchingBertModel
from torch.nn import MarginRankingLoss


from relogic.logickit.base.constants import *

from relogic.logickit.loss_func import get_loss
from torch.utils.checkpoint import checkpoint
import numpy as np
import torch
from functools import reduce

class Inference(nn.Module):
  """
  The inference module is to
    1. Run the inference process
    2. Calculate the loss
    3. Return the logits or loss depends on the mode (train or test)
  """
  def __init__(self, config, tasks):
    super(Inference, self).__init__()
    self.config = config
    self.tasks = tasks
    if config.branching_encoder:
      utils.log("Build Branching Bert Encoder")
      self.encoder = BranchingBertModel.from_pretrained(
        config.bert_model, encoder_structure=config.branching_structure)
    else:
      utils.log("Build {}:{} Encoder".format(config.encoder_type, config.bert_model))
      self.encoder = get_encoder(config.encoder_type).from_pretrained(
        config.bert_model,
        output_attentions=config.output_attentions)




    utils.log("Build Task Modules")
    self.tasks_modules = nn.ModuleDict()
    for task in tasks:
      if task.has_module:
        self.tasks_modules.update([(task.name, task.get_module())])
    self.task_dict = dict([(task.name, task) for task in self.tasks])
    self.dummy_input = torch.rand(1, 10, requires_grad=True)

    # self.encoder = HighwayLSTM(num_layers=3, input_size=300, hidden_size=200, layer_dropout=0.2)
    # self.word_embedding = nn.Embedding(self.config.external_vocab_size, self.config.external_vocab_embed_size)
    # self.word_embedding.weight.data.copy_(torch.from_numpy(np.load(config.external_embeddings)))
    # print("Loading embedding from {}".format(config.external_embeddings))

    self.loss_max_margin = MarginRankingLoss(margin=config.max_margin)
    self.distance = nn.PairwiseDistance(p=1)


  def encoding(self, **kwargs):
    task_name = kwargs.get("task_name", None)
    input_ids = kwargs.get("input_ids")
    input_mask = kwargs.get("input_mask")
    segment_ids = kwargs.get("segment_ids")
    extra_args = kwargs.get("extra_args", {})
    output_all_encoded_layers = extra_args.get("output_all_encoded_layers", False)
    route_path = extra_args.get("route_path", None)
    selected_non_final_layers = extra_args.get("selected_non_final_layers", None)
    no_dropout = task_name in READING_COMPREHENSION_TASKS
    langs = kwargs.get("langs", None)

    _input_token_ids = kwargs.get("_input_token_ids")
    _token_length = kwargs.get("_token_length")

    features = self.encoder(
      input_ids=input_ids,
      token_type_ids=segment_ids,
      attention_mask=input_mask,
      output_all_encoded_layers=output_all_encoded_layers,
      selected_non_final_layers=selected_non_final_layers,
      route_path=route_path,
      no_dropout=no_dropout,
      langs=langs,
      _input_token_ids=_input_token_ids,
      _token_length=_token_length)

    results = {}

    if isinstance(features, tuple):
      # Return Feature and Attention Map
      features, attention_map = features
      results["attention_map"] = attention_map

    if isinstance(features, list):
      # Return selected non final layers
      selected_non_final_layers_features = features[:-1]
      features = features[-1]
      results["selected_non_final_layers_features"] = selected_non_final_layers_features
    results["features"] = features

    return results

  def decoding(self, **kwargs):
    task_name = kwargs.get("task_name")
    logits = self.tasks_modules[task_name](**kwargs)
    return logits

  def get_arguments(self, prefix, kwargs):
    arguments = {}
    task_name = kwargs.pop("task_name", None)

    # LM features
    arguments["input_ids"] = kwargs.pop(prefix+"input_ids", None)
    arguments["input_mask"] = kwargs.pop(prefix+"input_mask", None)
    arguments["input_head"] = kwargs.pop(prefix+"input_head", None)
    arguments["segment_ids"] = kwargs.pop(prefix+"segment_ids", None)
    # arguments["label_ids"] = kwargs.get("label_ids", None)
    arguments["extra_args"] = kwargs.get("extra_args", {})
    arguments["output_all_encoded_layers"] = arguments["extra_args"].get("output_all_encoded_layers", False)
    arguments["route_path"] = arguments["extra_args"].get("route_path", None)
    arguments["selected_non_final_layers"] = arguments["extra_args"].get("selected_non_final_layers", None)
    arguments["no_dropout"] = task_name in READING_COMPREHENSION_TASKS

    # Classic features
    prefix = "_" + prefix
    arguments["_input_token_ids"] = kwargs.pop(prefix+"input_token_ids", None)
    arguments["_token_length"] = kwargs.pop(prefix+"token_length", None)
    # arguments["_label_ids"] = kwargs.pop(prefix+"_label_ids", None)
    return arguments


  def forward(self, *inputs, **kwargs):
    task_name = kwargs.pop("task_name")

    outputs_dict = {}

    # parallel teacher student task will use its own outputs_dict
    # It will iterate on each task, and get its output from decoder,
    # and save it on the `results`. The key is the task_name and value
    # is the tensor.
    if task_name == PARALLEL_TEACHER_STUDENT_TASK:
      teacher_prediction = kwargs.pop("teacher_predictions", None)
      if teacher_prediction is None:
        # Run teacher Part
        results = {}
        arguments = self.get_arguments(prefix="a_", kwargs=kwargs)
        encoding_results = self.encoding(**arguments)
        for task in self.tasks:
          if task.name != task_name:
            result = encoding_results["features"]
            results[task.name] = result.detach()# result.detach()
        return results

      else:
        arguments = self.get_arguments(prefix="b_", kwargs=kwargs)
        encoding_results = self.encoding(**arguments)

        losses = {}
        for task in self.tasks:
          if task.name != task_name:
            result = encoding_results["features"]
            logits, target, mask = self.decoding(
              task_name=task_name,
              student_results=result,
              teacher_results=teacher_prediction[task.name],
              **arguments, **kwargs)
            loss = get_loss(
              task=self.task_dict[task_name],
              logits=logits,
              config=self.config,
              target=target,
              mask=mask,
              **arguments, **kwargs)
            losses[task.name] = loss
        return reduce(lambda x,y:x+y, losses.values()), None

    elif task_name in SIAMESE:
      outputs_dict[task_name] = self.siamese_forward(*inputs, **kwargs, task_name=task_name)

    elif task_name == MIXSENT_TASK:
      outputs_dict[task_name] = self.mixsent_forward(*inputs, **kwargs, task_name=task_name)

    elif task_name in TRIPLET:
      outputs_dict[task_name] = self.triplet_forward(*inputs, **kwargs, task_name=task_name)

      # elif task_name in [DEP_PARSING_TASK]:
      #   arguments = self.get_arguments(prefix="", kwargs=kwargs)
      #   features = self.encoding(**arguments)
      #   outputs = self.decoding(**arguments, **kwargs, features=features, task_name=task_name)
      #   outputs_dict[task_name] = outputs

      # elif task_name == ENCODING_TASK:
      #   # For encoding task, for now, we only consider singleton.
      #   arguments = self.get_arguments(prefix="", kwargs=kwargs)
      #   features = self.encoder(**arguments)
      #   return {"features": features}

    else:
      # Currently we can only one-forward-multitask here
      # And the interface is not compatible with dependency parsing and some other multi-outputs task.
      # TODO: Restructure the inference to fix
      arguments = self.get_arguments(prefix="", kwargs=kwargs)
      encoding_results = self.encoding(**arguments)
      features = encoding_results.pop("features")
      kwargs.pop("extra_args", None)
      task_names = task_name.split(',')
      for task_name in task_names:
        logits = self.decoding(**arguments, **kwargs, features=features, task_name=task_name,
                               encoding_results=encoding_results)
        label_ids = kwargs.get("label_ids", None)
        _label_ids = kwargs.get("_label_ids", None)
        if label_ids is not None or _label_ids is not None:
          if task_name not in SKIP_LOSS_TASK:
          # if task_name in ["joint_srl"]:
          #   loss = self.task_dict[task_name].compute_loss()
          # else:
            loss = get_loss(
              task=self.task_dict[task_name],
              logits=logits,
              config=self.config,
              **arguments, **kwargs)
            outputs_dict[task_name] = {
              "loss": loss,
              "logits": logits}
          else:
            outputs_dict[task_name] = {"logits": logits}
        else:
          if task_name in SKIP_LOSS_TASK:
            outputs_dict[task_name] = {"logits": logits}
          else:
            outputs_dict[task_name] = {"logits": logits.detach()}
        if self.config.output_features:
          outputs_dict[task_name]["features"] = logits
      if self.config.output_attentions:
        outputs_dict[task_name]["attention_map"] = encoding_results["attention_map"]


    return outputs_dict


    # input_ids = kwargs.pop("input_ids")
    # input_mask = kwargs.pop("input_mask")
    # input_head = kwargs.pop("input_head", None)
    # segment_ids = kwargs.pop("segment_ids")
    # label_ids = kwargs.pop("label_ids", None)
    # extra_args = kwargs.pop("extra_args", {})
    # output_all_encoded_layers = extra_args.get("output_all_encoded_layers", False)
    # route_path = extra_args.get("route_path", None)
    # selected_non_final_layers = extra_args.get("selected_non_final_layers", None)
    # no_dropout = task_name in READING_COMPREHENSION_TASKS

    # BERT encoding
    # features = self.encoder(
    #   input_ids=input_ids,
    #   token_type_ids=segment_ids,
    #   attention_mask=input_mask,
    #   output_all_encoded_layers=output_all_encoded_layers,
    #   selected_non_final_layers=selected_non_final_layers,
    #   route_path=route_path,
    #   no_dropout=no_dropout)

    # features = checkpoint(self.encoder, {
    #   "input_ids": input_ids,
    #   "token_type_ids": segment_ids,
    #   "attention_mask": input_mask,
    #   "output_all_encoded_layers": output_all_encoded_layers,
    #   "selected_non_final_layers": selected_non_final_layers,
    #   "route_path": route_path,
    #   "no_dropout": no_dropout,
    # })

    # # LSTM encoder
    # input_token_ids = self.word_embedding(kwargs.pop("_input_token_ids"))
    # token_lengths = kwargs.pop("_token_length")
    # label_ids = kwargs.pop("_label_ids")
    # features = self.encoder(input_token_ids, token_lengths)

    # Semi-supervised Learning is not supported for now
    # There is a lot of semi-supervised learning algorithms,
    #  which are needed to be explored first, to make this
    #  framework more general

    # For each task, the interface is static
    #  including input_features, input_mask, segment_ids and extra_args

    # logits = self.tasks_modules[task_name](
    #   features = features,
    #   input_mask=input_mask,
    #   segment_ids=segment_ids,
    #   extra_args=extra_args,
    #   **kwargs)

    # logits = checkpoint(self.tasks_modules[task_name], {
    #   "features": features,
    #   "input_mask": input_mask,
    #   "segment_ids": segment_ids,
    #   "extra_args": extra_args,
    # }.update(kwargs))

    # if label_ids is not None:
    #   # if task_name in ["joint_srl"]:
    #   #   loss = self.task_dict[task_name].compute_loss()
    #   # else:
    #   loss = get_loss(
    #     task=self.task_dict[task_name],
    #     logits=logits,
    #     label_ids=label_ids,
    #     input_head=input_head,
    #     config=self.config,
    #     extra_args=extra_args,
    #     **kwargs)
    #   return loss, logits
    # else:
    #   return logits.detach()

  def siamese_forward(self, *inputs, **kwargs):
    task_name = kwargs.pop("task_name")

    a_arguments = self.get_arguments(prefix="a_", kwargs=kwargs)
    b_arguments = self.get_arguments(prefix="b_", kwargs=kwargs)

    a_encoding_results = self.encoding(**a_arguments)
    b_encoding_results = self.encoding(**b_arguments)
    a_features = a_encoding_results["features"]
    b_features = b_encoding_results["features"]

    logits = self.decoding(
      task_name=task_name,
      a_features=a_features,
      b_features=b_features,
      a_encoding_results=a_encoding_results,
      b_encoding_results=b_encoding_results,
      **kwargs)

    # loss = (logits * logits).sum() / logits.size(0)
    label_ids = kwargs.get("label_ids", None)
    _label_ids = kwargs.get("_label_ids", None)

    outputs_dict = {}

    if label_ids is not None or _label_ids is not None:
      if task_name not in SKIP_LOSS_TASK:
        loss = get_loss(
          task=self.task_dict[task_name],
          logits=logits,
          config=self.config,
          a_arguments=a_arguments,
          b_arguments=b_arguments,
          **kwargs)
        outputs_dict = {
          "loss": loss,
          "logits": logits}
      else:
        outputs_dict = {"logits": logits}
    else:
      if task_name in SKIP_LOSS_TASK:
        outputs_dict = {"logits": logits}
      else:
        outputs_dict = {"logits": logits.detach()}

    return outputs_dict

  def triplet_forward(self, *inputs, **kwargs):
    task_name = kwargs.pop("task_name")
    input_ids = kwargs.pop("input_ids")
    input_mask = kwargs.pop("input_mask")
    segment_ids = kwargs.pop("segment_ids")
    p_input_ids = kwargs.pop("p_input_ids")
    p_input_mask = kwargs.pop("p_input_mask")
    p_segment_ids = kwargs.pop("p_segment_ids")
    n_input_ids = kwargs.pop("n_input_ids")
    n_input_mask = kwargs.pop("n_input_mask")
    n_segment_ids = kwargs.pop("n_segment_ids")

    extra_args = kwargs.pop("extra_args", {})
    output_all_encoded_layers = extra_args.get("output_all_encoded_layers", False)
    route_path = extra_args.get("route_path", None)
    selected_non_final_layers = extra_args.get("selected_non_final_layers", None)

    is_inference = kwargs.pop("is_inference")

    encoding_results = self.encoder(
      input_ids=input_ids,
      token_type_ids=segment_ids,
      attention_mask=input_mask,
      output_all_encoded_layers=output_all_encoded_layers,
      selected_non_final_layers=selected_non_final_layers,
      route_path=route_path)

    features = encoding_results["features"]

    logits = self.tasks_modules[task_name](features)

    if not is_inference:
      p_encoding_results = self.encoder(
        input_ids=p_input_ids,
        token_type_ids=p_segment_ids,
        attention_mask=p_input_mask,
        output_all_encoded_layers=output_all_encoded_layers,
        selected_non_final_layers=selected_non_final_layers,
        route_path=route_path)

      n_encoding_results = self.encoder(
        input_ids=n_input_ids,
        token_type_ids=n_segment_ids,
        attention_mask=n_input_mask,
        output_all_encoded_layers=output_all_encoded_layers,
        selected_non_final_layers=selected_non_final_layers,
        route_path=route_path)

      p_features = p_encoding_results["features"]
      n_features = n_encoding_results["features"]

      p_logits = self.tasks_modules[task_name](p_features)
      n_logits = self.tasks_modules[task_name](n_features)

      positive_distance = self.distance(logits, p_logits)
      negative_distance = self.distance(logits, n_logits)
      loss = self.loss_max_margin(positive_distance, negative_distance, target= extra_args["target"])

      return {"loss": loss}
    else:
      return {"logits": logits}


  def mixsent_forward(self, *inputs, **kwargs):
    task_name = kwargs.pop("task_name")
    teacher_prediction = kwargs.pop("teacher_predictions", None)
    if teacher_prediction is None:
      # Run teacher Part
      results = {}
      arguments_a = self.get_arguments(prefix="a_", kwargs=kwargs)
      a_encdoing_results = self.encoding(**arguments_a)
      a_features = a_encdoing_results["features"]
      arguments_b = self.get_arguments(prefix="b_", kwargs=kwargs)
      b_encoding_results = self.encoding(**arguments_b)
      b_features = b_encoding_results["features"]
      for task in self.tasks:
        if task.name != task_name:
          # if isinstance(a_features, list):
          #   features_a = features_a[0]
          #   features_b = features_b[0]
          # else:
          #   features = features
          result_a = self.decoding(**arguments_a, **kwargs, features=a_features, task_name=task.name)
          result_b = self.decoding(**arguments_b, **kwargs, features=b_features, task_name=task.name)
          # TODO: Detach !!!
          results[task.name] = {
            "a": result_a.detach(),
            "b": result_b.detach()
          }
      return results

    else:
      # Run student and compute the loss
      arguments = self.get_arguments(prefix="c_", kwargs=kwargs)
      encoding_results = self.encoding(**arguments)
      features = encoding_results["features"]
      losses = {}
      for task in self.tasks:
        if task.name != task_name:
          result = self.decoding(**arguments, **kwargs, features=features, task_name=task.name)
          logits, target, mask = self.decoding(
            task_name=task_name,
            student_results=result,
            teacher_results=teacher_prediction[task.name],
            **arguments, **kwargs)
          loss = get_loss(
            task=self.task_dict[task_name],
            logits=logits,
            config=self.config,
            target=target,
            mask=mask,
            **arguments, **kwargs)
          losses[task.name] = loss
      return {"loss": reduce(lambda x,y: x+y, losses.values())}